#!/usr/bin/perl -w
########################################################
# apt-get install libtimedate-perl libnetwork-ipv4addr-perl libconfig-inifiles-perl
########################################################
# $event_hash -> $saved_IP_hash{'aa_' | ssh_saved_IPs.txt betöltése
# $event_hash -> @event_table         | ssh_syslog_lines.txt betöltése
# $event_hash -> @event_table         | syslog feldolgozása
#
# @event_table -> IP_type_hash        | ssh_syslog_lines visszaírásakor
# IP_type_hash -> $saved_IP_hash{'aa_'|'bb_' új banok
# $saved_IP_hash -> $banned_IP_hash   | amik még nem jártak le
########################################################

use strict;
use warnings;
use utf8;
use Encode qw( from_to encode decode );
use FileHandle;
use IO::Socket::UNIX qw( SOCK_STREAM SOMAXCONN );
use Thread::Queue;
use POSIX qw(:sys_wait_h);
use File::Basename;
use Net::IPv4Addr qw( ipv4_in_network );
use Sys::Syslog;
use Data::Dumper;
use Date::Parse;
use Date::Language;
use Config::IniFiles;
use Cwd 'abs_path';

my $prg = basename( $0 );

my $config_file       = "/etc/wwm.cfg";
my $local_config_file = "/etc/wwm.local.cfg";
my $work_dir = '';
my $trigger_time = 0;
my $unbanned_IP_remember_time = 0;
my $ban_time_multiplier = 0;
my $ipt_messages_repeate_time = 0;
my $ipt_errors_repeate_time = 0;
my $max_current_banspool_IP_to_syslog = 0;
my $logfile = '';
my $todo_config_file = '';
my $saved_IP_file = '';
my $saved_IP_newfile = '';
my $infile = '';
my $outfile = '';
my $log_data_file = '';
my %config_hash;
my @event_table;
my %saved_IP_hash;
my %banned_IP_hash;
my %IP_type_hash;
my %unbanned_IP_hash;
my $iptables_match_file_path = '';
my $iptables_recent_name = '';
my $iptables_recent_dir = '';
my $iptables_rules_method = '';
my $iptables_rule_chain = '';
my $iptables_rule = '';
my $iptables_rule_position_in_chain = '';
my $current_line = '';
my $banned_ip_list = '';
my $DATA = undef;
my $LOGDATA = undef;
my $lang = Date::Language->new('English');
my $just_started = 1;
my $current_time = 0;
my ( $outdated_ittems, $new_ittems, $new_ip, $banned_IP_count, $found_matching_lines)  =( 0, 0, 0, 0, 0 ) ;
my ($saved_inode, $saved_filepos, $new_saved_filepos, $log_inode ) = (0, 0, 0, 0);

#############################################################################
#############################################################################

sub _from_wide( $ ) {
  my $p0 = shift;
  my $p1 = '';
#  print "$p0   #\n";
  foreach my $item ( split( //, $p0 ) ) {
#    print "$item|| ".ord( $item )."\n";
    $p1 = $p1 . chr( ord( $item ) & 255 );
    $p1 = $p1 . chr( ( ord( $item ) / 256 ) & 255 );
  }
  from_to( $p1, 'utf16-le', 'utf8' );
  return $p1;
}

sub _to_utf8( $ ) {
  my $p0 = shift;
  my $p1 = '';
  my $p2 = '';
  my $p3 = '';
  my $p4 = '';
  foreach my $chars ( 'utf8', 'iso-8859-2', 'iso-8859-1' ) {
    $p1 = decode( $chars, $p0 );
    $p2 = encode( 'utf8', $p1 );
    $p3 = decode( 'utf8', $p2 );
    $p4 = encode( $chars, $p3 );
    if ( $p1 eq $p3 and $p2 ne $p3 and $p4 eq $p0 ) {
      if (    ( $p4 ne $p2 and $chars =~ /^iso/ )
           or ( $p4 ne $p3 and $chars =~ /^utf/ ) ) {
        return $p2;
      }
    }
  }
  return $p0;
}

sub print_to_queue {
  my $tmp_str = shift;
  foreach my $line (split /^/, $tmp_str) {
    if ( $line =~ m/'(EXPIRE|TIMESTAMP)' => (\d+)(,|)$/ ) {
#       print "1 $1 2 $2 3 $3";
      chomp($line);
      my $lang_date = Date::Language->new('English');
      my $decoded_time = $lang_date->time2str( '%Y.%m.%d %H:%M:%S', $2);
      $line .= " ($decoded_time)\n";
    }
    print _to_utf8( _from_wide( $line ) );
  }
}

sub logit { ##################### sub logit - Begin
  my( $msg ) = @_;
  if( length( $msg ) == 0 ) {
    return 1;
  }
  $msg = _to_utf8( _from_wide( $msg ) );
  print_to_queue "###-Logit-### $msg\n";
#  setlogsock( 'unix' ); # http://perldoc.perl.org/Sys/Syslog.html  [THE RULES OF SYS::SYSLOG]
  openlog( $prg, 'pid,ndelay,nowait', 'user' );
  syslog( 'info', $msg );
  closelog(  );
  return 1;
} ################################ sub logit - End

sub show_changes_in_config { ##################### sub show_changes_in_config - Begin
#  my ($a, $b) = @_;  
  return;
} ################################ sub show_changes_in_config - End

sub recalculate_expire_times { ##################### sub recalculate_expire_times - Begin
  my ($a, $b) = @_;
  my $have_old_config = 0;
  my $sect = 'basic';
  my $param = '';
  if ( ! ref $a && ref($a) ne 'HASH' ) {
    print_to_queue "recalculate_expire_times \$a is not a HASH reference\n";
    return;
  }
  if ( ! ref $b && ref($b) ne 'HASH' ) {
    print_to_queue "recalculate_expire_times \$b is not a HASH reference\n";
    return;
  }
  if ( keys ( % {$a} ) > 0 ) {
    $have_old_config = 1;
  }
  # event_table lejárati idők újraszámolása, ha kell
  $param = 'trigger_time';
  if ( $have_old_config && (! defined $a->{$sect} || ! defined $a->{$sect}{$param} ) ) {
    print_to_queue "recalculate_expire_times \$a->{$sect}{$param} not defined\n";
    return;
  }
  if ( ! defined $b->{$sect} || ! defined $b->{$sect}{$param} ) {
    print_to_queue "recalculate_expire_times \$b->{$sect}{$param} not defined\n";
    return;
  }
  if ( ! $have_old_config || $a->{$sect}{$param} != $b->{$sect}{$param} ) {
    print_to_queue "recalculate_expire_times: trigger_time changed from $a->{$sect}{$param} to $b->{$sect}{$param}\n" if $have_old_config;
    my $updated_expire_time = 0;
    foreach (@event_table) {
      my $TimeStamp = $_->{TIMESTAMP};
      my $Expire = $_->{EXPIRE};
      my $NewExpire = $TimeStamp + $b->{$sect}{$param};
      if ( $have_old_config && ( $TimeStamp + $a->{$sect}{$param} != $Expire ) ) {
        print_to_queue "recalculate_expire_times WARNING: expire time of event [$_->{COMMENT}] from IP [$_->{IP}] didn't match caclulated time\n";
        print_to_queue "recalculate_expire_times found: [$Expire] should be: [".($TimeStamp + $a->{$sect}{$param})."]\n";
      }
      if ( $NewExpire != $Expire ) {
        $_->{EXPIRE} = $NewExpire;
        $updated_expire_time++;
      }
    }
    print_to_queue "recalculate_expire_times: $updated_expire_time \@event_table expire time(s) updated\n" if $updated_expire_time;
  }
  # saved_IP_hash lejárati idők újraszámolása, ha kell
  $param = 'ban_time_multiplier';
  if ( $have_old_config && (! defined $a->{$sect} || ! defined $a->{$sect}{$param} ) ) {
    print_to_queue "recalculate_expire_times \$a->{$sect}{$param} not defined\n";
    return;
  }
  if ( ! defined $b->{$sect} || ! defined $b->{$sect}{$param} ) {
    print_to_queue "recalculate_expire_times \$b->{$sect}{$param} not defined\n";
    return;
  }
  my %tmp_hash;
  $tmp_hash{'repeated'} = 0;
  foreach (keys %saved_IP_hash) {
    push @{ $tmp_hash{'event_types'}{ $saved_IP_hash{$_}{'TYPE'} }{'keys'} }, $_;
    if ( defined $saved_IP_hash{$_}{'REPEATED'} ) {
      $tmp_hash{'repeated'} = 1;
    }
  }
  if ( $tmp_hash{'repeated'} == 1 && (! $have_old_config || $a->{$sect}{$param} != $b->{$sect}{$param} ) ) {
    print_to_queue "recalculate_expire_times: $param changed from $a->{$sect}{$param} to $b->{$sect}{$param}\n" if $have_old_config;
  }
  foreach my $E_type( keys %{ $tmp_hash{'event_types'} } ) {
    $sect = 'events';
    $param = 'ban_time';
    my $updated_expire_time = 0;
    if ( $have_old_config && (! defined $a->{$sect} || ! defined $a->{$sect}{$E_type} || ! defined $a->{$sect}{$E_type}{$param} ) ) {
      print_to_queue "recalculate_expire_times \$a->{$sect}{$E_type}{$param} not defined\n";
      return;
    }
    if ( ! defined $b->{$sect} || ! defined $b->{$sect}{$E_type} || ! defined $b->{$sect}{$E_type}{$param} ) {
      print_to_queue "recalculate_expire_times \$b->{$sect}{$E_type}{$param} not defined\n";
      return;
    }
    if ( ! $have_old_config || $a->{$sect}{$E_type}{$param} != $b->{$sect}{$E_type}{$param} ) {
      print_to_queue "recalculate_expire_times: $param changed for event $E_type from $a->{$sect}{$E_type}{$param} to $b->{$sect}{$E_type}{$param}\n" if $have_old_config;
      if ( $have_old_config && $a->{$sect}{$E_type}{'type'} ne $b->{$sect}{$E_type}{'type'} ) {
        print_to_queue "recalculate_expire_times WARNING: 'type' changed for event $E_type from $a->{$sect}{$E_type}{$param} to $b->{$sect}{$E_type}{$param}\n";
      }
      foreach ( @{ $tmp_hash{'event_types'}{$E_type}{'keys'} } ) {
        my $TimeStamp = $saved_IP_hash{$_}{TIMESTAMP};
        my $SavedExpire = $saved_IP_hash{$_}{EXPIRE};
        my $repeated = $saved_IP_hash{$_}{REPEATED} || 0;
        my $old_ban_multip = $a->{'basic'}{'ban_time_multiplier'};
        my $new_ban_multip = $b->{'basic'}{'ban_time_multiplier'};
        my $NewExpire = $TimeStamp;
        my $OldExpire = $TimeStamp;
        if ( $repeated ) {
          $NewExpire += $b->{$sect}{$E_type}{$param} * $new_ban_multip * $repeated;
          $OldExpire += $a->{$sect}{$E_type}{$param} * $old_ban_multip * $repeated;
        } else {
          $NewExpire += $b->{$sect}{$E_type}{$param};
          $OldExpire += $a->{$sect}{$E_type}{$param};
        }
        if ( $have_old_config && ( $OldExpire != $SavedExpire ) ) {
          print_to_queue "recalculate_expire_times WARNING: expire time of event [".$saved_IP_hash{$_}{COMMENT}."] from IP [$saved_IP_hash{$_}{IP}] didn't match caclulated time\n";
          print_to_queue "recalculate_expire_times found: [$SavedExpire] should be: [$OldExpire]\n";
        }
        if ( $NewExpire != $SavedExpire ) {
          $saved_IP_hash{$_}{EXPIRE} = $NewExpire;
          $updated_expire_time++;
        }
      }
      print_to_queue "recalculate_expire_times: $updated_expire_time %saved_IP_hash expire times for event $E_type updated\n" if $updated_expire_time;
    }
  }
  # unbanned_IP_hash lejárati idők újraszámolása, ha kell
  $sect = 'basic';
  $param = 'unbanned_IP_remember_time';
  if ( $have_old_config && (! defined $a->{$sect} || ! defined $a->{$sect}{$param} ) ) {
    print_to_queue "recalculate_expire_times \$a->{$sect}{$param} not defined\n";
    return;
  }
  if ( ! defined $b->{$sect} || ! defined $b->{$sect}{$param} ) {
    print_to_queue "recalculate_expire_times \$b->{$sect}{$param} not defined\n";
    return;
  }
  if ( ! $have_old_config || $a->{$sect}{$param} != $b->{$sect}{$param} ) {
    print_to_queue "recalculate_expire_times: $param changed from $a->{$sect}{$param} to $b->{$sect}{$param}\n" if $have_old_config;
    my $updated_expire_time = 0;
    foreach (keys %unbanned_IP_hash) {
      my $TimeStamp = $unbanned_IP_hash{$_}{TIMESTAMP};
      my $Expire = $unbanned_IP_hash{$_}{EXPIRE};
      my $NewExpire = $TimeStamp + $b->{$sect}{$param};
      if ( $have_old_config && ( $TimeStamp + $a->{$sect}{$param} != $Expire ) ) {
        print_to_queue "recalculate_expire_times WARNING: expire time of event [".$unbanned_IP_hash{$_}{COMMENT}."] from IP [$_] didn't match caclulated time\n";
        print_to_queue "recalculate_expire_times found: [$Expire] should be: [".($TimeStamp + $a->{$sect}{$param})."]\n";
      }
      if ( $NewExpire != $Expire ) {
        $unbanned_IP_hash{$_}{EXPIRE} = $NewExpire;
        $updated_expire_time++;
      }
    }
    print_to_queue "recalculate_expire_times: $updated_expire_time %unbanned_IP_hash expire times updated\n" if $updated_expire_time;
  }
  return;
} ################################ sub recalculate_expire_times - End

sub read_value_from_config {
  my( $config, $section, $param ) = @_;
  my @tmp = $config->val( $section,  $param );
  if (@tmp > 1) {
    print_to_queue "WARNING! Multiline config parameter [$section][$param]\n";
    my $first_line = shift @tmp;
    my $identical = 1;
    foreach ( @tmp ) {
      if ( $first_line ne $_ ) {
        $identical--;
        last;
      }
    }
    if ( $identical ) {
      print_to_queue "WARNING! Lines appears to be identical. Ignoring copies. Parameter value will be:\n";
      print_to_queue "[".$first_line."]\n";
      return $first_line;
    } else {
      print_to_queue "WARNING! Lines appears to be different. They merged to a single line:\n";
      print_to_queue "[".$first_line.join('', @tmp)."]\n";
      return $first_line.join('', @tmp);
    }
  }
  return $tmp[0];
}

sub read_config {
  my $reason = shift;
  my $print_lines = 0;
  my $print_final_config = 0;
  my $print_reload_changes = 1;
  my %new_config_hash;
  my @configs = ( { type => 'system', file => ( -l $config_file )?abs_path($config_file):$config_file },
                  { type => 'local',  file => ( -l $local_config_file )?abs_path($local_config_file):$local_config_file },
                );
  foreach ( @configs ) {
    if ( -f $_->{'file'} ) {
      my $config_type = $_->{'type'};
      my $config_file = $_->{'file'};
      print_to_queue "loading $_->{'type'} config_file = $config_file\n";
      my $config = Config::IniFiles->new( -file => $config_file );
#print "---------- $config_file -------------------------------\n". Dumper($config) . "\n\n";
      if (! defined $config ) {
        logit "Config file '$config_file' has errors:\n".join('\n',@Config::IniFiles::errors);
        if ( $reason eq 'initial' ) {
          exit 102;
        } else { return 0; }
      }
      my $sect = 'include';
      foreach my $param ( $config->Parameters($sect) ) {
        my $include_file = read_value_from_config ($config, $sect, $param );
        $include_file = abs_path($include_file) if -l $include_file;
        my $uniq = 1;
        foreach ( @configs ) {
          if ($_->{'file'} eq $include_file) {
            $uniq = 0;
            last;
          }
        }
        if ( $uniq ) {
          push @configs, { type => 'local', file => $include_file };
          logit "Include file [$include_file] in parameter [$param] in config file [$config_file] '$sect' section added to the end of the loadable config file list";
        } else {
          logit "Include file [$include_file] in parameter [$param] is not uniq in config file [$config_file] '$sect' section!";
        }
      }
      $sect = 'basic';
      my %known_params;
      # A 'white_networks' paraméterrel trükközni kell, hogy ne írja felül a tömb referenciát
      foreach ( qw ( trigger_time unbanned_IP_remember_time ban_time_multiplier ipt_messages_repeate_time ipt_errors_repeate_time logfile work_dir max_current_banspool_IP_to_syslog supress_white_network_messages) ) {
        $known_params{$_}{'have'} = 0;
      }
      foreach my $param ( $config->Parameters($sect) ) {
        my $tmp = read_value_from_config ($config, $sect, $param );
        if ( defined $known_params{$param} ) {
          $new_config_hash{$sect}{$param} = $tmp;
          $known_params{$param}{'have'} = 1;
          print_to_queue "$param = $tmp\n" if $print_lines;
        } elsif ($param ne 'white_networks') {
          logit "Unknown parameter [$param] ignored in config file [$config_file] '$sect' section!";
        }
      }
      foreach ( keys %known_params ) {
        if ( $known_params{$_}{'have'} == 0 && $config_type eq 'system') {
          logit "Mandatory parameter [$_] missing from config file [$config_file] '$sect' section!";
          exit 101;
        }
      }
      {
        my $tmp = read_value_from_config ($config, $sect, 'white_networks' );
        if ( defined $tmp ) {
          my @tmpa = split(/;/, $tmp );
          my @white_networks;
          for (my $ii = 0 ;  $ii < @tmpa ; $ii++) {
            my ($ip,$cidr) = eval { Net::IPv4Addr::ipv4_parse( $tmpa[$ii] ) };
            if ( !defined $ip || !defined $cidr ) {
              logit "'white_networks' parameter in config file [$config_file] '$sect' section must contain valid IP address range(s)! [$tmpa[$ii]] IP/mask(;IP/mask)*";
              next;
            }
            push @white_networks, $tmpa[$ii];
          }
          if ( $config_type eq 'system') {
            $new_config_hash{'basic'}{'white_networks'} = \@white_networks;
          } else {
            my $array_ref = $new_config_hash{'basic'}{'white_networks'};
            while ( @{ $array_ref } ) { pop @{ $array_ref }; }
            push @{ $array_ref }, @white_networks;
          }
        } elsif ( $config_type eq 'system') {
          logit "Mandatory parameter [white_networks] missing from config file [$config_file] '$sect' section!";
          exit 101;
        }
      }
      $sect = 'iptables';
      foreach ( qw ( iptables_rules_method iptables_recent_name iptables_rule_chain iptables_rule_position_in_chain) ) {
        $known_params{$_}{'have'} = 0;
      }
      foreach my $param ( $config->Parameters($sect) ) {
        my $tmp = read_value_from_config ($config, $sect, $param );
        if ( defined $known_params{$param} ) {
          $new_config_hash{$sect}{$param} = $tmp;
          $known_params{$param}{'have'} = 1;
          print_to_queue "$param = $tmp\n" if $print_lines;
        } else {
          logit "Unknown parameter [$param] ignored in config file [$config_file] '$sect' section!";
        }
      }
      foreach ( keys %known_params ) {
        if ( $known_params{$_}{'have'} == 0 && $config_type eq 'system') {
          logit "Mandatory parameter [$_] missing from config file [$config_file] '$sect' section!";
          exit 101;
        }
      }
      if ( $new_config_hash{'iptables'}{'iptables_rules_method'} !~ m/^create_rule|maintain_ips$/ ) {
        logit "Parameter [iptables_rules_method] value in config file [$config_file] '$sect' section is invalid!";
        logit "  value must be 'create_rule' or 'maintain_ips'!";
        if ( $reason eq 'initial' ) {
          exit 101;
        } else { return 0; }
      }
      if ( $new_config_hash{'iptables'}{'iptables_rules_method'} eq 'maintain_ips' &&
           $new_config_hash{'iptables'}{'iptables_recent_name'} eq '' ) {
        logit "Parameter [iptables_recent_name] value in config file [$config_file] '$sect' section must not be empty!";
        logit "  value must be an existing iptables recent rule name!";
        if ( $reason eq 'initial' ) {
          exit 101;
        } else { return 0; }
      }
      if ( $new_config_hash{'iptables'}{'iptables_rules_method'} eq 'create_rule' &&
           $new_config_hash{'iptables'}{'iptables_rule_chain'} eq '' ) {
        logit "Parameter [iptables_rule_chain] value in config file [$config_file] '$sect' section must not be empty!";
        logit "  value must be a valid iptables chain name!";
        if ( $reason eq 'initial' ) {
          exit 101;
        } else { return 0; }
      }
      if ( $new_config_hash{'iptables'}{'iptables_rules_method'} eq 'create_rule' &&
           ( $new_config_hash{'iptables'}{'iptables_rule_position_in_chain'} eq '' ||
             $new_config_hash{'iptables'}{'iptables_rule_position_in_chain'} !~ m/^\d+$/)
         ) {
        logit "Parameter [iptables_rule_position_in_chain] value in config file [$config_file] '$sect' section must be a positive number!";
        logit "  value must be a positive number which defines the position of the created iptables rule position in $iptables_rule_chain chain!";
        if ( $reason eq 'initial' ) {
          exit 101;
        } else { return 0; }
      }
      $sect = 'events';
      foreach my $param ( $config->Parameters($sect) ) {
        if ( $param !~ m/^(\d+)$/ ) {
          logit "Parameter names in config file [$config_file] '$sect' section must be number!";
          next;
        }
        my ( $num, $value ) = ($1, read_value_from_config ($config, $sect, $param ) );
        if ( $value !~ m/^(B|W)``(\d+)``(\d+)``(.*)$/ ) {
          logit "Parameter [$param] value in config file [$config_file] '$sect' section did not match /(B|W)``\\d+``\\d+``.*/.";
          next;
        }
        $new_config_hash{'events'}{$num}{'type'}          = $1;
        $new_config_hash{'events'}{$num}{'trigger_count'} = $2 * 1;
        $new_config_hash{'events'}{$num}{'ban_time'}      = $3 * 1;
        $new_config_hash{'events'}{$num}{'friendly_name'} = $4;
        print_to_queue "$num =  $value\n" if $print_lines;
      }
      print_to_queue "#event total: ". keys( %{$new_config_hash{'events'}} ) ."\n" if $print_lines;

      foreach $sect ( $config->Sections ) {
        if ( $sect !~ m/^proto-(.*)/ ) { next; }
        my $proto_name = $1;
        print_to_queue "proto =  $proto_name\n" if $print_lines;
        my $proto_prefix = read_value_from_config ($config, $sect, 'prefix' );
        if ( defined $proto_prefix ) {
          my $validate = eval { qr/$proto_prefix/ };
#          print_to_queue "validate= $validate; $@\n" if $print_lines;
          if ( $@ ) {
            logit "'prefix' regexp value in config file [$config_file] '$sect' section is invalid:";
            logit "$@";
            if ( $config_type eq 'system' ) {
              if ( $reason eq 'initial' ) {
                exit 101;
              } else { return 0; }
            } else {
              { next; } # foreach $sect ( $config->Sections )
            }
          }
          $new_config_hash{'proto'}{$proto_name}{'prefix'} = $proto_prefix;
          print_to_queue "  prefix =  $proto_prefix\n" if $print_lines;
        } elsif ( !defined $new_config_hash{'proto'}{$proto_name} || 
                  !defined $new_config_hash{'proto'}{$proto_name}{'prefix'} ) {
          logit "'prefix' value in config file [$config_file] '$sect' section is not defined!";
          if ( $config_type eq 'system' ) {
            if ( $reason eq 'initial' ) {
              exit 101;
            } else { return 0; }
          } else {
            { next; } # foreach $sect ( $config->Sections )
          }
        }
        foreach my $param ( $config->Parameters($sect) ) {
          if ( $param =~ m/^suffixes_([a-z]+)_(\d+)$/ ) {
            my ($a1, $a2 ) = ($1, $2 );
            if ( !defined $new_config_hash{'proto'}{$proto_name}{'suffixes_'.$a1} ) {
# nem szabad ezt: $new_config_hash{'proto'}{$proto_name}{'suffixes_'.$a1}{'fields'} vizsgálni, mert a fasz perl -1 elemig létrehozza az üres hasht!
              next;
            }
            my $value = read_value_from_config ($config, $sect, $param );
            if ( $value !~ m/^(\d+)``(.*)$/ ) {
              logit "Parameter [$param] value in config file [$config_file] '$sect' section did not match /\\d+``<regexp>/.";
              next;
            }
            my ($p_type, $p_pattern) = ($1, $2);
            if ( ! defined $new_config_hash{'events'}{$p_type} ) {
              logit "Parameter [$param] value in config file [$config_file] '$sect' section is invalid.";
              logit "Event type [$p_type] did not exist!";
              next;
            }
            if ( $p_pattern =~ m/``/ ) {
              logit "Parameter [$param] value in config file [$config_file] '$sect' section is invalid.";
              logit "Pattern [$p_pattern] part should not contain duble backticks ``!";
              next;
            }
            my $validate = eval { qr/$p_pattern/ };
            if ( $@ ) {
              logit "Parameter [$param] value in config file [$config_file] '$sect' section is invalid.";
              logit "Pattern [$p_pattern] has regex errors:";
              logit "$@";
              if ( $config_type eq 'system' ) {
                if ( $reason eq 'initial' ) {
                  exit 101;
                } else { return 0; }
              } else {
                { next; }
              }
            }
            $new_config_hash{'proto'}{$proto_name}{'suffixes_'.$a1}{'patterns'}{$a2}{'type'}     = $p_type;
            $new_config_hash{'proto'}{$proto_name}{'suffixes_'.$a1}{'patterns'}{$a2}{'pattern'}  = $p_pattern;
            print_to_queue "  suffixes_".$a1."_$a2 =  $value\n" if $print_lines;
          }
          if ( $param =~ m/^suffixes_[a-z]+$/ ) {
            my $value = read_value_from_config ($config, $sect, $param );
            if ( $value !~ m/^((dummy|username)``)*IP(``(dummy|username))*$/ ) {
              logit "Parameter [$param] value in config file [$config_file] '$sect' section did not match /((dummy|username)``)*IP(``(dummy|username))*/.";
              logit "Patterns of '$param' group will not be loaded and used";
            } elsif ( $value =~ m/username.*username/ ) {
              logit "Parameter [$param] value in config file [$config_file] '$sect' section have more than one username field. It must have only one.";
              logit "Patterns of '$param' group will not be loaded and used";
            } else {
              $new_config_hash{'proto'}{$proto_name}{$param}{'fields'} = $value;
              print_to_queue "  $param =  $value\n" if $print_lines;
            }
          }
        } # foreach my $param ( $config->Parameters($sect) )
        foreach my $white_param ( qw (white_networks white_network_add) ) {
          my $config_val_white_networks = read_value_from_config ($config, $sect, $white_param );
          if ( defined $config_val_white_networks ) {
            my @tmpa = split(/;/, $config_val_white_networks );
            my @white_networks;
            for (my $ii = 0 ;  $ii < @tmpa ; $ii++) {
              my ($ip,$cidr) = eval { Net::IPv4Addr::ipv4_parse( $tmpa[$ii] ) };
              if ( !defined $ip || !defined $cidr ) {
                logit "Parameter [$white_param] value in config file [$config_file] '$sect' section contain valid IP address range(s)!";
                logit "[$tmpa[$ii]] part is not valid and will be ignored.";
                next;
              }
              push @white_networks, $tmpa[$ii];
            }
            $new_config_hash{'proto'}{$proto_name}{$white_param} =  \@white_networks;
          } elsif ($white_param eq 'white_networks' ) {
            $new_config_hash{'proto'}{$proto_name}{$white_param} =  \@{ $new_config_hash{'basic'}{$white_param} };
          }
        }
      } # foreach $sect ( $config->Sections )
    } else {
      logit "Config file ($_->{'file'}) did not exist!";
      if ( $_->{'type'} eq 'system' ) {
        if ( $reason eq 'initial' ) {
          exit 101;
        } else { return 0; }
      }
    }
  } # foreach ( @configs )
  print_to_queue "read_config ($reason) Done.\n";
  recalculate_expire_times(\%config_hash, \%new_config_hash);
  if ( $print_reload_changes ) {
    show_changes_in_config(%config_hash, %new_config_hash);
  }
  for (keys %config_hash) { delete $config_hash{$_}; }
  ###################################################################################################
  %config_hash = %new_config_hash; # ################################################################
  $iptables_recent_name = $new_config_hash{'iptables'}{'iptables_recent_name'};
  $iptables_rules_method = $new_config_hash{'iptables'}{'iptables_rules_method'};
  $iptables_rule_chain = $new_config_hash{'iptables'}{'iptables_rule_chain'};
  $iptables_rule_position_in_chain = $new_config_hash{'iptables'}{'iptables_rule_position_in_chain'};
  $iptables_rule = "-p all -m recent --name $iptables_recent_name --rcheck -j DROP";
  $trigger_time                      = $new_config_hash{'basic'}{'trigger_time'};
  $unbanned_IP_remember_time         = $new_config_hash{'basic'}{'unbanned_IP_remember_time'};
  $ban_time_multiplier               = $new_config_hash{'basic'}{'ban_time_multiplier'};
  $ipt_messages_repeate_time         = $new_config_hash{'basic'}{'ipt_messages_repeate_time'};
  $ipt_errors_repeate_time           = $new_config_hash{'basic'}{'ipt_errors_repeate_time'};
  $max_current_banspool_IP_to_syslog = $new_config_hash{'basic'}{'max_current_banspool_IP_to_syslog'};
  $logfile                           = $new_config_hash{'basic'}{'logfile'};
  $work_dir                          = $new_config_hash{'basic'}{'work_dir'};
  $todo_config_file                  = $work_dir . "todo";
  $saved_IP_file                     = $work_dir . "saved_IPs.txt";
  $saved_IP_newfile                  = $work_dir . "saved_IPs.txt.new";
  $infile                            = $work_dir . "saved_syslog_lines.txt";
  $outfile                           = $work_dir . "saved_syslog_lines.txt.new";
  $log_data_file                     = $work_dir . "logdata.txt";
  {
    $Data::Dumper::Purity = 1;
    $Data::Dumper::Sortkeys = sub {
      my @KEYS = keys %{$_[0]};
      return [@KEYS] unless defined $KEYS[0];
      $KEYS[0] =~ m/^\d+$/ ? return [sort {$a <=> $b} @KEYS ] : return [sort {$a cmp $b} @KEYS ] 
    };
    my $new_config_hash = { %config_hash };
    print_to_queue '' . Data::Dumper->Dump([$new_config_hash], ['new_config_hash']) if $print_final_config;
  }
  return 1;
}

sub _is_ipv4 {
  my $value = shift;
  my ( @octets ) = $value =~ /^(\d{1,3})\.(\d{1,3})\.(\d{1,3})\.(\d{1,3})$/;
  return unless ( @octets == 4 );
  foreach ( @octets ) {
    return unless ( $_ >= 0 && $_ <= 255 );
  }
  return 1;
}

sub put_in_IP_type_hash { ##################### sub put_in_IP_type_hash - Begin
  my $element = shift;
  my $e_type = $element->{'TYPE'};
  my $e_IP = $element->{'IP'};
  my $e_repeated = $element->{'REPEATED'} || 0;
  if ( $e_repeated ) {
    $IP_type_hash{$e_IP}{$e_type}{'REPEATED'} = $element->{'REPEATED'};
  }
  $IP_type_hash{$e_IP}{$e_type}{'count'} += 1;
  $IP_type_hash{$e_IP}{$e_type}{'TIMESTAMP'} = $element->{'TIMESTAMP'};
  $IP_type_hash{$e_IP}{$e_type}{'COMMENT'} = $element->{'COMMENT'};
  return 1;
} ############################################# sub put_in_IP_type_hash - End

sub check_iptables_match_file_path { ############################## sub check_iptables_match_file_path - Begin
  $iptables_match_file_path = '';
  $iptables_recent_dir = '';
  foreach my $dir ( qw (ipt_recent xt_recent) ) {
    if (-f "/proc/net/$dir/$iptables_recent_name" ) {
      $iptables_match_file_path = "/proc/net/$dir/$iptables_recent_name";
      $iptables_recent_dir = $dir;
      last;
    }
  }
  if ( $iptables_match_file_path eq '' ) {
    logit ("[aa] configured 'iptables_recent_name' didn't exist: /proc/net/(ipt_recent|xt_recent)/$iptables_recent_name didn't exist!");
    return 0;
  }
  return 1;
} ############################################# sub check_iptables_match_file_path - End

sub change_ip_list_tot { ############################## sub change_ip_list_tot - Begin
  my ($ipt_list_tot_ref, $ipt_count, $change_to) = @_;
  if ( $$ipt_list_tot_ref < $change_to &&
      ( !defined $config_hash{'ipt'}{'ipt_count_near_full'} ||
                 $config_hash{'ipt'}{'ipt_count_near_full'} == 0 )
     ) {
    $config_hash{'ipt'}{'ipt_count_near_full'} = $current_time;
    logit ("ipt_count is near full: ".$$ipt_list_tot_ref."/$ipt_count" );
  }
  if ( $$ipt_list_tot_ref > $change_to &&
      ( !defined $config_hash{'ipt'}{'ipt_count_should_lower'} ||
                 $config_hash{'ipt'}{'ipt_count_should_lower'} == 0 )
     ) {
    $config_hash{'ipt'}{'ipt_count_should_lower'} = $current_time;
    logit ("ipt_count can be lowered: ".$$ipt_list_tot_ref."/$ipt_count" );
  }
  my $change_ipt_count_result = `chmod 644 /sys/module/$iptables_recent_dir/parameters/ip_list_tot && echo $change_to > /sys/module/$iptables_recent_dir/parameters/ip_list_tot`;
  if ( $change_ipt_count_result ne '' ) {
    if ( !defined $config_hash{'ipt'}{'change_ipt_count_result'} || $config_hash{'ipt'}{'change_ipt_count_result'} == 0 ) {
      $config_hash{'ipt'}{'change_ipt_count_result'} = $current_time;
      logit ("ipt_count change was not successfull: $change_ipt_count_result");
    }
  } else {
    $$ipt_list_tot_ref = $change_to;
    logit ("ipt_count changed to: $$ipt_list_tot_ref");
  }
} ############################################# sub change_ip_list_tot - End

sub ipt_recent { ############################## sub ipt_recent - Begin
  my ($operation, $ip) = @_;
  if ( $operation !~ m/^add|remove|clear$/) {
    logit ("[aa] syntax error in 'ipt_recent' call operation= [$operation] != [add|remove|clear] ");
    return undef;
  }
  my $prefix = '';
  if    ( $operation eq 'add' )    { $prefix = '+' }
  elsif ( $operation eq 'remove' ) { $prefix = '-' }
  elsif ( $operation eq 'clear' )  { $prefix = '/' }
  if ( check_iptables_match_file_path() ) {
    my $ipt_list_tot = `cat /sys/module/$iptables_recent_dir/parameters/ip_list_tot`;
    my $ipt_count = `cat $iptables_match_file_path | wc -l`;
    chomp $ipt_list_tot;
    if ( $ipt_list_tot !~ m/^(\d+)$/ ) {
      if ( !defined $config_hash{'ipt'}{'ipt_list_tot_not_numeric'} || $config_hash{'ipt'}{'ipt_list_tot_not_numeric'} == 0 ) {
        $config_hash{'ipt'}{'ipt_list_tot_not_numeric'} = $current_time;
        logit ("ipt_list_tot is not numeric: $ipt_list_tot");
      }
      $ipt_list_tot = 0;
    }
    chomp $ipt_count;
    if ( $ipt_count !~ m/^(\d+)$/ ) {
      if ( !defined $config_hash{'ipt'}{'ipt_count_not_numeric'} || $config_hash{'ipt'}{'ipt_count_not_numeric'} == 0 ) {
        $config_hash{'ipt'}{'ipt_count_not_numeric'} = $current_time;
        logit ("ipt_count is not numeric: $ipt_count");
      }
      $ipt_count = 0;
    }
    if ( $operation eq 'add' && $ipt_count + 20 > $ipt_list_tot ) {
      change_ip_list_tot(\$ipt_list_tot, $ipt_count, $ipt_list_tot+20 );
    }
    if ( $operation eq 'remove' && $ipt_list_tot >= 120 && $ipt_count < $ipt_list_tot - 40 ) {
      change_ip_list_tot(\$ipt_list_tot, $ipt_count, $ipt_list_tot-20);
    }
    my $result = `echo $prefix$ip > $iptables_match_file_path`;
    chomp($result);
    if ($result ne '' ) {
      logit ("[aa] error in iptables command: [echo $prefix$ip > $iptables_match_file_path]: $result");
      return undef;
    } else { return 1; }
  } else {
    if ( !defined $config_hash{'ipt'}{'ipt_recent_file_not_found'} ||
                   $config_hash{'ipt'}{'ipt_recent_file_not_found'} == 0 ) {
      $config_hash{'ipt'}{'ipt_count_should_lower'} = $current_time;
      logit ("[aa] configured 'iptables_recent_name' didn't exist: /proc/net/(ipt_recent|xt_recent)/$iptables_recent_name didn't exist!");
    }
    
    return undef;
  }
} ############################################# sub ipt_recent - End

sub iptables_check_rule { ############################## sub iptables_check_rule - Begin
  my $print_match = shift || 0;
  if ( $iptables_rules_method eq 'create_rule' ) {
    my $cmd_result = `iptables -C $iptables_rule_chain $iptables_rule 2>&1`;
    if ( $cmd_result ne '' ) {
      # nincs találat
      logit "Banning rule did not exist in iptables '$iptables_rule_chain' chain :[$cmd_result]!";
      if ( $just_started ) {
        logit "There is no point staying resident if no banning rule present! Quiting.";
        exit 200;
      }
      return 0;
    } else {
      print_to_queue "matching rule:[$iptables_rule_chain $iptables_rule]\n" if ( $print_match );
      return 1;
    }
  } else {
    my @cmd_result = `iptables -S | grep $iptables_recent_name 2>&1`;
    my @grep_result = grep(/--name\s+$iptables_recent_name/, @cmd_result);
    chomp @grep_result;
    if ( scalar @grep_result == 0 ) {
      logit "Banning rule did not exist in any iptables filter chain (no match for '--name $iptables_recent_name')!";
      if ( $just_started ) {
        logit "There is no point staying resident if no banning rule present! Quiting.";
        exit 200;
      }
      return 0;
    } else {
      print_to_queue "first matching rule:[$grep_result[0]]\n" if ( $print_match );
      return 1;
    }
  }
} ############################################# sub iptables_check_rule - End

sub iptables_create_rule { ############################## sub iptables_create_rule - Begin
  my $cmd_result = `iptables -I $iptables_rule_chain $iptables_rule_position_in_chain $iptables_rule 2>&1`;
  if ( $cmd_result ne '' ) {
    # sikertelen generálás
    $cmd_result =~ s/^iptables: (.*)$/$1/;
    logit "Error inserting ban rule to '$iptables_rule_chain' chain at position $iptables_rule_position_in_chain :[$cmd_result]!";
    if ( $cmd_result eq 'Index of insertion too big.' ) {
      logit "Will try to append the rule to the given chain but it may not worth a thing!";
      `iptables -A $iptables_rule_chain $iptables_rule 2>&1`;
    }
  } else {
    print_to_queue "iptables rule creation was successful\n";
  }
} ############################################# sub iptables_create_rule - End

sub iptables_delete_rule { ############################## sub iptables_delete_rule - Begin
  my $cmd_result = `iptables -D $iptables_rule_chain $iptables_rule 2>&1`;
  if ( $cmd_result ne '' ) {
    # sikertelen generálás
    $cmd_result =~ s/^iptables: (.*)$/$1/;
    logit "Error deleting ban rule from '$iptables_rule_chain' chain :[$cmd_result]!";
  } else {
    print_to_queue "iptables rule deletion was successful\n";
  }
} ############################################# sub iptables_delete_rule - End

my $flag = "/var/run/$prg";
if (-f $flag ) {
  my $tmpstr = "ps awxf | grep '$prg' | grep perl | grep -v grep | awk '{print " . '$1' . "}'";
  my $run_check = `$tmpstr`;
  if ( $run_check ne '' ) {
    logit( "Already running" );
    sleep 1;
    exit 0;
  }
}
my $F1;
die "Semaphore file ($flag) creation failed: $!" unless open $F1, '>', $flag;
die "Semaphore file ($flag) close failed: $!"unless close $F1;

STDERR->autoflush(1);
STDOUT->autoflush(1);
read_config('initial');
print_to_queue "working directory: [$work_dir]\n";
#print "---------- config_hash -------------------------------\n". Dumper(%config_hash) . "\n\n";
#exit 0;
iptables_create_rule() if $iptables_rules_method eq 'create_rule';
iptables_check_rule(1); # will quit if no banning rule present

###################### mentett file-ok megnyitasa ###################################
{
  my $start_time = time();
  my $SAVED_IP_FILE = undef;
  if (-f $saved_IP_file) {
    $SAVED_IP_FILE = FileHandle->new($saved_IP_file);
    if (!defined $SAVED_IP_FILE) {
      unlink( $flag );
      die "Unable to open SAVED_IP_FILE ($saved_IP_file) file: $!\n";
    }
  }
  my $BEMENET = undef;
  if (-f $infile) {
    $BEMENET = FileHandle->new($infile);
    if (!defined $BEMENET) {
      unlink( $flag );
      die "Unable to open BEMENET ($infile) file: $!\n";
    }
  }
######################## mentett IP fa'jl (ssh_saved_IPs.txt) feldolgoza'sa ####################################
  if (defined $SAVED_IP_FILE) {
    print_to_queue "Loading saved events from file ($saved_IP_file).\n";
    my $lines = 0;
    my $input_errors = 0;
    my $expired_lines = 0;
    my $banned_ip = 0;
    my $white_ip = 0;
    while (<$SAVED_IP_FILE>) {
      chomp;
      $lines++;
      if ( $_ eq '' ) { next; }
      if ( m/^#/ ) { next; }
      if ( m/^([^!]+)!([^!]+)!([^!]+)!([^!]+)$/ ) {
        my ($a1, $a2, $a3, $a4 ) = ($1,$2,$3,$4);
        my %event_hash = ();
        if (!( $a1 =~ m/\d+/ )) {
          print_to_queue "Invalid TIMESTAMP value [$a1] at line $lines [$_]\n";
          $input_errors++;
          next;
        }
        $event_hash{'TIMESTAMP'} = $a1 * 1;
        if ( ! defined _is_ipv4( $a2 ) ) {
          print_to_queue "Invalid IP value [$a2] at line $lines [$_]\n";
          $input_errors++;
          next;
        }
        $event_hash{'IP'} = $a2;
        my $repe_ated = 0;
        if ( $a3 =~ m/^R(\d+)(.*)/ ) { # Repeated event
          $a3 = $1;
          if ( $2 =~ m/^-(\d+)-(\d+)$/ ) {
            $event_hash{'REPEATED'} = $1;
          } else {
            $event_hash{'REPEATED'} = 1;
          }
          $event_hash{'EXPIRE'} = $a1 + $config_hash{'events'}{ $a3 }{'ban_time'} * $ban_time_multiplier * $event_hash{'REPEATED'};
        }
        if (! defined $config_hash{'events'}{$a3} ) {
          print_to_queue "Invalid TYPE value [$a3] at line $lines [$_]\n";
          $input_errors++;
          next;
        }
        $event_hash{'TYPE'} = $a3;
        if ( $config_hash{'events'}{ $event_hash{'TYPE'} }{'type'} eq 'B' ) {
          $banned_ip++;
        } elsif ( $config_hash{'events'}{ $event_hash{'TYPE'} }{'type'} eq 'W' ) {
          $white_ip++;
        }
        $event_hash{'EXPIRE'} = $a1 + $config_hash{'events'}{ $a3 }{'ban_time'} if !defined $event_hash{'REPEATED'};
        $event_hash{'COMMENT'} = $a4;
        if ( $start_time > $event_hash{'EXPIRE'} ) {
          $expired_lines++;
           if ( $config_hash{'events'}{ $event_hash{'TYPE'} }{'type'} eq 'B' ) {
            my $commited = ipt_recent('remove', $a2);
            print_to_queue "IP [$a2] removed from $iptables_match_file_path (expired)\n" if $commited;
          }
         
        } else {
          $saved_IP_hash{'aa_'.$lines}= \%event_hash;
          if ( !defined $banned_IP_hash{ $a2 } && $config_hash{'events'}{ $event_hash{'TYPE'} }{'type'} eq 'B' ) {
            my $commited = ipt_recent('add', $a2);
            print_to_queue "IP [$a2] reinserted to $iptables_match_file_path\n" if $commited;
          }
          $banned_IP_hash{ $a2 }{'TIMESTAMP'} = $event_hash{'TIMESTAMP'};
        }
      } else {
        print_to_queue "Invalid syntax at line $lines [$_]\n";
        $input_errors++;
      }
    }
    close ($SAVED_IP_FILE);
    print_to_queue "Found $lines lines in saved IP-s file ($saved_IP_file). Errors in file = $input_errors. Expired items = $expired_lines.\n";
    print_to_queue "Banned IPs: $banned_ip. White IPs: $white_ip\n";
  }
######################### bemeneti fa'jl (ssh_syslog_lines.txt) feldolgoza'sa ###################################
  if (defined $BEMENET) { 
    print_to_queue "Loading saved loglines from file ($infile).\n";
    my $lines = 0;
    my $input_errors = 0;
    my $expired_lines = 0;
    my $ban_type_line = 0;
    my $white_type_line = 0;
    while (<$BEMENET>) {
      chomp;
      if ( $_ eq '' ) { next; }
      if ( m/^#/ ) { next; }
      $lines++;
      if ( m/^([^!]+)!([^!]+)!([^!]+)!([^!]+)$/ ) {
        my ($a1, $a2, $a3, $a4 ) = ($1,$2,$3,$4);
        my %event_hash = ();
        if (  ! defined _is_ipv4( $a2 ) ) {
          print_to_queue "Invalid IP value [$a2] at line $lines [$_]\n";
          $input_errors++;
          next;
        }
        $event_hash{'IP'} = $a2;
        if (!( $a1 =~ m/\d+/ )) { 
          print_to_queue "Invalid TIMESTAMP value [$a1] at line $lines [$_]\n";
          $input_errors++;
          next;
        }
        $event_hash{'TIMESTAMP'} = $a1 * 1;
        if (! defined $config_hash{'events'}{$a3} ) {
          print_to_queue "Invalid TYPE value [$a3] at line $lines [$_]\n";
          $input_errors++;
          next;
        }
        $event_hash{'TYPE'} = $a3;
        if ( $config_hash{'events'}{ $event_hash{'TYPE'} }{'type'} eq 'B' ) {
          $ban_type_line++;
        } elsif ( $config_hash{'events'}{ $event_hash{'TYPE'} }{'type'} eq 'W' ) {
          $white_type_line++;
        }
        $event_hash{'EXPIRE'} = $a1 + $config_hash{'basic'}{'trigger_time'};
        $event_hash{'COMMENT'} = $a4;
        if ( $start_time > $event_hash{'EXPIRE'} ) {
          $expired_lines++;
        } else {
          push @event_table , \%event_hash;
        }
      } else {
        print_to_queue "Invalid syntax at line $lines [$_]\n";
        $input_errors++;
      }
    }
    close ($BEMENET);
    print "Found $lines lines in saved loglines file ($infile). Errors in file = $input_errors. Expired items = $expired_lines.\n";
    print_to_queue "Ban type saved lines: $ban_type_line. White type saved lines: $white_type_line\n";
  }
################################# log fa'jl (syslog) feldolgoza'sa ############################################
  $LOGDATA = FileHandle->new($log_data_file, "r");
  if (defined $LOGDATA) {
    if ( $LOGDATA->getline =~ m/^([^!]+)!([^!]+)$/ ) {
      $saved_inode = $1 *1;
      $saved_filepos = $2 *1;
    } else {
      print "hibas a log_data_file formatuma\n";
    }
    close ($LOGDATA);
  }
}
print_to_queue "--------------------------------------------------------------------------\n";
print_to_queue "--------------------------------------------------------------------------\n";
logit( "Started" );

###############################################################################################################
################################# folyamatos ciklus kezdete ###################################################
###############################################################################################################
###############################################################################################################
my $tmp_str;
my $sleep_time = 1;
while (1) {
if ( -f $todo_config_file ) {
  my $saved_todo_config_file = $todo_config_file; # kell, hogy ha konfigban wordir-t változtat akkor ne haljon bele
  my $config = FileHandle->new($saved_todo_config_file);
  if (!defined $config) { die "Unable to open todo_config_file ($saved_todo_config_file) file: $!\n"; }
# print_to_queue "============= Fri Apr  3 13:38:14 2015 ======================================================\n";
  print_to_queue "=============================================================================================\n";
  print_to_queue "============= ".scalar localtime().  " ========== todo_config_file exist ====================\n";
  while (<$config>) {
    $current_line = $_;
    chomp($current_line);
    next if $current_line eq '';
    if ($current_line eq 'exit') {
      print_to_queue "============= exit\n";
      logit( "Exit" );
      unlink( $flag );
      if (!unlink ($saved_todo_config_file) ) {
        print_to_queue "Unable to delete todo_config_file ($saved_todo_config_file) file: $!\n";
      }
      exit 100;
    }
    if ($current_line eq 'unload') {
      print_to_queue "============= unload\n";
      logit( "Unload" );
      unlink( $flag );
      if (!unlink ($saved_todo_config_file) ) {
        print_to_queue "Unable to delete todo_config_file ($saved_todo_config_file) file: $!\n";
      }
      my $commited = ipt_recent('clear', '');
      print_to_queue "All IP was deleted from $iptables_match_file_path\n" if $commited;;
      iptables_delete_rule() if $iptables_rules_method eq 'create_rule';
      exit 100;
    }
    if ($current_line eq 'restart') {
      print_to_queue "============= restart\n";
      logit( "Restart" );
      unlink( $flag );
      if (!unlink ($saved_todo_config_file) ) {
        print_to_queue "Unable to delete todo_config_file ($saved_todo_config_file) file: $!\n";
      }
      exit 101;
    }
    if ($current_line eq 'reload_config') { # ###############################################
      print_to_queue "============= reload_config\n";
      logit( "Reload config" );
      read_config('reload_config');
      next;
    }
    if ($current_line eq 'dump-config') { # ###############################################
      print_to_queue "============= dump-config\n";
      my $new_config_hash = { %config_hash };
      print_to_queue Dumper($new_config_hash)."\n================================\n";
      next;
    }
    if ($current_line eq 'dump-basic') { # ###############################################
      print_to_queue "============= dump-basic\n";
      foreach ( sort keys %{$config_hash{'basic'} } ) {
        print_to_queue "  $_: ".$config_hash{'basic'}{$_}."\n";
      }
      next;
    }
    if ($current_line eq 'dump-iptables') { # ###############################################
      print_to_queue "============= dump-iptables\n";
      foreach ( sort keys %{$config_hash{'iptables'} } ) {
        print_to_queue "  $_: ".$config_hash{'iptables'}{$_}."\n";
      }
      next;
    }
    if ($current_line eq 'dump-events') { # ###############################################
      print_to_queue "============= dump-events\n";
      my $new_config_hash = { %{$config_hash{'events'} } };
      print_to_queue Dumper($new_config_hash)."\n================================\n";
      next;
    }
    if ($current_line eq 'dump-banned_ip_list') { # ###############################################
      print_to_queue "============= dump-banned_ip_list\n";
      print_to_queue "$banned_ip_list\n================================\n";
      next;
    }
    if ($current_line eq 'dump-white_networks') { # ###############################################
      print_to_queue "============= dump-white_networks\n";
      my $basic_white_ref = $config_hash{'basic'}{'white_networks'};
      print_to_queue "  == global: ".join(';', @{ $basic_white_ref })."\n";
      foreach my $proto ( keys %{$config_hash{'proto'}} ) {
        my $proto_white_ref = $config_hash{'proto'}{$proto}{'white_networks'};
        if ( $proto_white_ref == $basic_white_ref ) {
          print_to_queue "  == $proto: <inherited from global> ".join(';', @{ $proto_white_ref })."\n";
        } else {
          print_to_queue "  == $proto: ".join(';', @{ $proto_white_ref })."\n";
        }
        if ( defined $config_hash{'proto'}{$proto}{'white_network_add'} ) {
          print_to_queue "    == + ".join(';', @{ $config_hash{'proto'}{$proto}{'white_network_add'} })."\n";
        }
      }
      print_to_queue "================================\n";
      next;
    }
    if ($current_line =~ m/^add-(ban|white) (.*)/) { # ###############################################
      my $maual_event_type = $1;
      my $ip_address = $2;
      my ($friendly_name, $event_num ,$event_type);
      if ( $maual_event_type eq 'ban' ) {
        $friendly_name = 'manual ban';
        $event_num = 17;
        $event_type = 'B';
      } elsif ( $maual_event_type eq 'white' ) {
        $friendly_name = 'manual whitelist';
        $event_num = 16;
        $event_type = 'W';
      }
      if ( !defined $config_hash{'events'}{$event_num} || $config_hash{'events'}{$event_num}{'type'} ne $event_type ) {
        print_to_queue ("  == Event '$event_num' type is not '$event_type'! Default config altered! Not adding event!");
        next;
      }
      if ( $config_hash{'events'}{$event_num}{'friendly_name'} ne $friendly_name ) {
        print_to_queue ("  == Event '$event_num' friendly_name is not '$friendly_name'! Default config altered! Not adding event!");
        next;
      }
      if ( $config_hash{'events'}{$event_num}{'trigger_count'} != 0 ) {
        print_to_queue ("  == Event '$event_num' trigger_count is not '0'! Default config altered! Not adding event!");
        next;
      }
      my $event_time = $config_hash{'events'}{$event_num}{'ban_time'};
      if ( $ip_address =~ m/^(.*) (.*)/) {
        $ip_address = $1;
        $event_time = $2 * 1;
      }
      if ( !defined _is_ipv4( $ip_address ) ) {
        print_to_queue ("  == The IP field didn't contain a vaid IPv4 address!");
        print_to_queue ("  == [$ip_address]\n");
        next;
      }
      $current_time = time();
      $saved_IP_hash{'mm_'.$event_type.'_'.$ip_address}{'IP'} = $ip_address;
      $saved_IP_hash{'mm_'.$event_type.'_'.$ip_address}{'TIMESTAMP'} = $current_time;
      $saved_IP_hash{'mm_'.$event_type.'_'.$ip_address}{'TYPE'} = $event_num;
      $saved_IP_hash{'mm_'.$event_type.'_'.$ip_address}{'EXPIRE'} = $current_time + $event_time;
      $saved_IP_hash{'mm_'.$event_type.'_'.$ip_address}{'COMMENT'} = $friendly_name;
      if ( $maual_event_type eq 'white' ) {
      logit ("[a] New event [$friendly_name] for user [N/A] from IP [$ip_address] added to event spool for $event_time seconds");
      } elsif ( $maual_event_type eq 'ban' ) {
        my $commited = ipt_recent('add', $ip_address);
        logit ("[a] New event [$friendly_name] from IP [$ip_address] added to iptables recent list for $event_time seconds") if $commited;
      }
      $just_started = 1; # hogy a változások megtörténjenek azonnal
      next;
    }
    if ($current_line =~ m/^delete-banned-IP (.*)/) { # ###############################################
      my $search_IP = $1;
      my $ii = 0;
      print_to_queue "delete-banned-IP ($search_IP)\n";
      for ( my $event_table_array_item_count = scalar( @event_table ); $event_table_array_item_count > 0; $event_table_array_item_count-- ) {
        my $element = shift @event_table;
        if ( $element->{'IP'} ne $search_IP ) {
          # visszatesszük a végére, ha nem kell törölni
          push @event_table, $element;
        } else {
          $ii++;
          my $event_type = $element->{'TYPE'};
          if ( $config_hash{'events'}{ $event_type }{'type'} eq 'B'  ) { # ha ban
            print_to_queue "  == event_table match found: '".$config_hash{'events'}{ $event_type }{'friendly_name'}."' (event_type: $event_type)\n";
            print_to_queue "    == $element->{'TIMESTAMP'}!$element->{'COMMENT'}\n";
            # felszabaditjuk a hasht
            %{ $element } = ();
          }
          if ( $config_hash{'events'}{ $event_type }{'type'} eq 'W'  ) { # ha white esemény
            # visszatesszük a végére
            push @event_table, $element;
            print_to_queue "  == event_table match found: '".$config_hash{'events'}{ $event_type }{'friendly_name'}."' (event_type: $event_type), not deleting white event!\n";
          }
        }
      }
      print_to_queue "  == event_table $ii match found\n" if $ii == 0;
      $ii = 0;
      foreach ( keys %saved_IP_hash ) {
        next if ( !defined $saved_IP_hash{$_}->{'IP'} || $saved_IP_hash{$_}->{'IP'} ne $search_IP );
        $ii++;
        my $event_type = $saved_IP_hash{$_}->{'TYPE'};
        if ( $config_hash{'events'}{ $event_type }{'type'} eq 'B'  ) { # ha ban
          print_to_queue "  == saved_IP_hash match found: '".$config_hash{'events'}{ $event_type }{'friendly_name'}."' (event_type: $event_type)\n";
          print_to_queue "    == ".$saved_IP_hash{$_}{'TIMESTAMP'}."!".$saved_IP_hash{$_}{'COMMENT'}."\n";
          delete $saved_IP_hash{$_};
        }
        if ( $config_hash{'events'}{ $event_type }{'type'} eq 'W'  ) { # ha white esemény
          print_to_queue "  == saved_IP_hash match found: '".$config_hash{'events'}{ $event_type }{'friendly_name'}."' (event_type: $event_type), not deleting white event!\n";
        }
      }
      print_to_queue "  == saved_IP_hash $ii match found\n" if $ii == 0;
      if ( defined $IP_type_hash{$search_IP} ) {
        foreach my $event_type ( keys %{$IP_type_hash{$search_IP}} ) {
          if ( $config_hash{'events'}{$event_type}{'type'} eq 'B'  ) { # ha ban
            print_to_queue "  == IP_type_hash match found: '".$config_hash{'events'}{$event_type}{'friendly_name'}."' (event_type: $event_type)\n";
            print_to_queue "    == ".$IP_type_hash{$search_IP}{$event_type}{'TIMESTAMP'}."!".$IP_type_hash{$search_IP}{$event_type}{'COMMENT'}."\n";
            delete $IP_type_hash{$search_IP}{$event_type};
          }
          if ( $config_hash{'events'}{ $event_type }{'type'} eq 'W'  ) { # ha white esemény
            print_to_queue "  == IP_type_hash match found: '".$config_hash{'events'}{ $event_type }{'friendly_name'}."' (event_type: $event_type), not deleting white event!\n";
          }
        }
      } else {
        print_to_queue "  == IP_type_hash NO match found\n";
      }
      my %tmp_hash0 = ( 'banned_IP_hash' => \%banned_IP_hash,
                        'unbanned_IP_hash' => \%unbanned_IP_hash,
                        );
      foreach my $hash_name ( %tmp_hash0 ) {
        next if not defined $tmp_hash0{$hash_name};
        my $hash_ref = $tmp_hash0{$hash_name};
        if ( defined $hash_ref->{$search_IP} ){
          delete $hash_ref->{$search_IP};
          print_to_queue "  == $hash_name match found\n";
        } else {
          print_to_queue "  == $hash_name NO match found\n";
        }
      }
      print_to_queue "  == checking if $search_IP is banned in $iptables_match_file_path\n";
      my $ipt_content = `cat $iptables_match_file_path | grep $search_IP`;
      chomp $ipt_content;
      if ( $ipt_content ne '' ) {
        my $commited = ipt_recent('remove', $search_IP );
        print_to_queue "    == $iptables_match_file_path match found\n";
        logit ("[b] IP [$search_IP] deleted from iptables recent file (todo_config_file)") if $commited;
      } else {
        print_to_queue "    == $iptables_match_file_path NO match found\n";
      }
      $just_started = 1; # hogy a változások megtörténjenek azonnal
      print_to_queue "  Done. Printing $search_IP IP matches from internal tables\n";
      $current_line = 'dump-IP '.$search_IP; # rögtön ki is írjuk az 'eredményt'
    }
    if ($current_line =~ m/^dump-IP (.*)/) { # ###############################################
      my $search_IP = $1;
      my $ii = 0;
      $Data::Dumper::Pad = '    ';
      print_to_queue "============= dump-IP $search_IP\n";
      print_to_queue "  == \@event_table ===\n";
      foreach ( @event_table ) {
        next if ( !defined $_->{'IP'} || $_->{'IP'} ne $search_IP );
        print_to_queue Dumper($_)."\n";
        $ii++;
      }
      print_to_queue "    NO match found\n" if $ii == 0;
      $ii = 0;
      print_to_queue "  == saved_IP_hash ==\n";
      foreach ( sort {$a cmp $b} keys %saved_IP_hash ) {
        next if ( !defined $saved_IP_hash{$_}->{'IP'} || $saved_IP_hash{$_}->{'IP'} ne $search_IP );
        my %tmp_hash2;
        $tmp_hash2{$_} = $saved_IP_hash{$_};
        my $dumped_data = { %tmp_hash2 };
        print_to_queue Dumper( $dumped_data )."\n";
        $ii++;
      }
      print_to_queue "    NO match found\n" if $ii == 0;
      my %tmp_hash0 = ( 'IP_type_hash'    => \%IP_type_hash,
                        'banned_IP_hash' => \%banned_IP_hash,
                        'unbanned_IP_hash' => \%unbanned_IP_hash,
                        );
      foreach ( %tmp_hash0 ) {
        next if not defined $tmp_hash0{$_};
        print_to_queue "  == $_ ==\n";
        my %tmp_hash1 = %{$tmp_hash0{$_}};
        $ii = 0;
        foreach ( sort {$a cmp $b} keys %tmp_hash1 ) {
          next if ( !defined $tmp_hash1{$_} || $_ ne $search_IP );
          my %tmp_hash2;
          $tmp_hash2{$_} = $tmp_hash1{$_};
          my $dumped_data = { %tmp_hash2 };
          print_to_queue Dumper( $dumped_data )."\n";
          $ii++;
        }
        print_to_queue "    NO match found\n" if $ii == 0;
      }
      print_to_queue "=============== end of dump-IP $search_IP =================\n";
      $Data::Dumper::Pad = '';
      next;
    }
    if ($current_line =~ m/^dump-\@event_table (.*)/) { # ###############################################
      my $ii = 0;
      my $ex_regex = $1;
      print_to_queue "============= dump-\@event_table $ex_regex".' total '.@event_table." element\n";
      foreach ( @event_table ) {
        $ii++;
        next if (defined $_->{'TYPE'} && $_->{'TYPE'} !~ m/$ex_regex/);
        print_to_queue $ii.' : '.Dumper($_)."\n";
      }
      print_to_queue "\n================================\n";
      next;
    }
    if ($current_line =~ m/^dump-proto (.*)/) { # ###############################################
      my $proto = $1;
      print_to_queue "============= dump-proto $proto\n";
      if ( defined $config_hash{'proto'}{$proto} ) {
        my $P_hash = $config_hash{'proto'}{$proto};
        print_to_queue "  prefix: ".$P_hash->{'prefix'}."\n";
        print_to_queue "  white_networks: ".join(';', @{ $P_hash->{'white_networks'} })."\n";
        foreach my $suffix_name ( sort keys %{$P_hash} ) {
          next if $suffix_name !~ m/^suffixes_([a-z]+)$/;
          print_to_queue "  $suffix_name fields: ".$P_hash->{$suffix_name}{'fields'}."\n";
          foreach my $suffix_number ( sort keys %{$P_hash->{$suffix_name}{'patterns'}} ) {
            print_to_queue "  $suffix_name\_$suffix_number type:".$P_hash->{$suffix_name}{'patterns'}{$suffix_number}{'type'}.
                   " pattern: ".$P_hash->{$suffix_name}{'patterns'}{$suffix_number}{'pattern'}."\n";
          }
        }
      }
      print_to_queue "================================\n\n";
      next;
    }
    if ($current_line =~ m/^dump-%saved_IP_hash (.*)/) { # ###############################################
      my $ii = 0;
      my $ex_regex = $1;
      print_to_queue "============= dump-%saved_IP_hash $ex_regex\n";
      foreach ( sort {$a cmp $b} keys %saved_IP_hash ) {
        $ii++;
        next if (defined $saved_IP_hash{$_}{'TYPE'} && $saved_IP_hash{$_}{'TYPE'} !~ m/$ex_regex/);
        my %tmp_hash;
        $tmp_hash{$_} = $saved_IP_hash{$_};
        my $dumped_data = { %tmp_hash };
        print_to_queue $ii.' : ['.$saved_IP_hash{$_}{'TYPE'}.'] '.Dumper( $dumped_data )."\n";
      }
      print_to_queue "\n================================\n";
      next;
    }
    if ($current_line eq 'dump-%saved_IP_hash') { # ###############################################
      print_to_queue "============= dump-%saved_IP_hash\n";
      my $dumped_data = { %saved_IP_hash };
      print_to_queue ''.Dumper($dumped_data)."\n================================\n";
      next;
    }
    if ($current_line eq 'dump-%IP_type_hash') {
      print_to_queue "============= dump-%IP_type_hash\n";
      my $dumped_data = { %IP_type_hash };
      print_to_queue ''.Dumper($dumped_data)."\n================================\n";
      next;
    }
    if ($current_line eq 'dump-%banned_IP_hash') {
      print_to_queue "============= dump-%banned_IP_hash\n";
      my $dumped_data = { %banned_IP_hash };
      print_to_queue ''.Dumper($dumped_data)."\n================================\n";
      next;
    }
    if ($current_line eq 'dump-%unbanned_IP_hash') {
      print_to_queue "============= dump-%unbanned_IP_hash\n";
      my $dumped_data = { %unbanned_IP_hash };
      print_to_queue ''.Dumper($dumped_data)."\n================================\n";
      next;
    }
    print_to_queue "Unknown todo_config_file command: $current_line (".scalar localtime().")\n";
  }
  if (!unlink ($saved_todo_config_file) ) {
    die "Unable to delete todo_config_file ($saved_todo_config_file) file: $! (".scalar localtime().")\n";
  }
}

# @event_table        nem szabad törölni, ebben van az összes eddig felolvasott esemény.
# %saved_IP_hash      nem szabad törölni, ebben van az összes eddig kiadott ban.
# %banned_IP_hash     ebben keletkezik a teljes ban lista, ha van új esemény.
# %IP_type_hash       innen kerülnek át a próbálkozók a %saved_IP_hash változóba, ha elégszer próbálkoztak már.

$current_time = time();
$DATA = undef;
$found_matching_lines = 0;

if ( -f $logfile) {
  $DATA = FileHandle->new($logfile);
  if (!defined $DATA) { die "Unable to open DATA ($logfile) file: $!\n"; }
  $log_inode = (stat $DATA)[1];
  if ($log_inode == $saved_inode) {
    my $log_size = (-s $logfile);
    if ( $log_size < $saved_filepos ) {
      print "log_size < saved_filepos ($log_size)<($saved_filepos)\n";
      $saved_filepos = $log_size;
    }
    seek $DATA, $saved_filepos, 0;
    #print "saved_filepos = $saved_filepos\n";
  } else {
      print_to_queue '######## '.scalar localtime()." -------\n";
      print_to_queue "logfile inode has changed (logrotate) ($saved_inode -> $log_inode)\n";
      $saved_inode = $log_inode;
      $saved_filepos = 0;
  }
} else { die "DATA ($logfile) file did not exist!\n$!\n"; }

while (<$DATA>) {
  chomp;
  my $debug_pattern_matching = 0;
  $current_line = $_;
  print_to_queue "**debug_pattern_matching** line: [$current_line]\n" if $debug_pattern_matching;
  my $found_match = 0;
  foreach my $proto ( keys %{$config_hash{'proto'}} ) {
    print_to_queue "**debug_pattern_matching** proto: [$proto]\n" if $debug_pattern_matching == 2;
    print_to_queue "**debug_pattern_matching** prefix: [".$config_hash{'proto'}{$proto}{'prefix'}."]\n" if $debug_pattern_matching == 2;
    if ( $current_line =~ m/$config_hash{'proto'}{$proto}{'prefix'}/ ) {
      my $date_time = $1;
      my $line_suffix = $current_line;
      $line_suffix =~ s/$config_hash{'proto'}{$proto}{'prefix'}//;
      print_to_queue "**debug_pattern_matching**   line_suffix: [$line_suffix]\n" if $debug_pattern_matching == 2;
      my $time_stamp = $lang->str2time( $date_time );
      if (!( $time_stamp =~ m/\d+/ )) {
        print_to_queue "The TIMESTAMP [$time_stamp] section of the current line [$current_line] is invalid.\n";
        next;
      }
      foreach my $suffix_name ( sort keys %{$config_hash{'proto'}{$proto}} ) {
        if ( $suffix_name !~ m/^suffixes_([a-z]+)$/ ) { next; }
        print_to_queue "**debug_pattern_matching**   suffix_name: [$suffix_name]\n" if $debug_pattern_matching == 2;
        my @fields_array = split(/``/, $config_hash{'proto'}{$proto}{$suffix_name}{'fields'});
        my %fields_hash = ();
        for ( my $qq = 0 ; $qq < @fields_array ; $qq++ ) { $fields_hash{ $fields_array[$qq] } = $qq; }
        if ( ! defined $fields_hash{'IP'} ) { # ugyan az elején levizsgálom, de hátha...
          logit ("There is no IP field marked in '$proto' section '$suffix_name' field value?!: [$config_hash{'proto'}{$proto}{$suffix_name}{'fields'}]");
          logit ("fields_hash:" . Dumper(%fields_hash));
          logit ("'$suffix_name' from '$proto' section will be deleted!");
          delete $config_hash{'proto'}{$proto}{$suffix_name};
          next;
        }
        foreach my $suffix_number ( sort keys %{$config_hash{'proto'}{$proto}{$suffix_name}{'patterns'}} ) {
          print_to_queue "**debug_pattern_matching**     suffix_number [$suffix_number]\n" if $debug_pattern_matching == 2;
          my $actual_pattern = $config_hash{'proto'}{$proto}{$suffix_name}{'patterns'}{$suffix_number}{'pattern'};
          if ( my @suffix_matches_array = $line_suffix =~ m/$actual_pattern/ ) {
            $found_matching_lines++; # kiíratáshoz és további feldolgozáshoz
            $found_match++; # pre és suffix vizsgálathoz. Ha volt találat nem kell a többit vizsgálnunk
            print_to_queue "**debug_pattern_matching**       Found match![$proto][$suffix_name\_$suffix_number]\n" if $debug_pattern_matching;
            my $username = 'N/A' ;
            if ( defined $fields_hash{'username'} ) {
              $username = $suffix_matches_array[ $fields_hash{'username'} ];
            }
            my $ip_address = $suffix_matches_array[ $fields_hash{'IP'} ];
            if ( !defined $ip_address ) {
              logit ("The IP marked field in the matched pattern is not defined! [$actual_pattern]");
              logit ("syslog line: [$line_suffix]");
              last;
            } elsif ( $ip_address eq '' ) {
              logit ("!! The IP marked field in the matched pattern is empty!");
              logit ("!! [$actual_pattern]");
              logit ("!! syslog line: [$line_suffix]");
              last;
            } elsif ( !defined _is_ipv4( $ip_address ) ) {
              print_to_queue ("!! The IP marked field in the matched pattern didn't contain a vaid IPv4 address!");
              print_to_queue ("!! [$actual_pattern]\n");
              print_to_queue ("!! syslog line: [$line_suffix]\n");
              last;
            }
            my @white_networks = @{ $config_hash{'proto'}{$proto}{'white_networks'} };
            push @white_networks, @{ $config_hash{'proto'}{$proto}{'white_network_add'} } if defined $config_hash{'proto'}{$proto}{'white_network_add'};
            my $is_lan_IP = 0;
            foreach ( @white_networks ) {
              if ( $_ eq '' ) { next; }
              if ( Net::IPv4Addr::ipv4_in_network ( $_, $ip_address ) ) {
                print_to_queue "IP ($ip_address) is in white network list ($_) for '$proto' ... skipping\n" if ! $config_hash{'basic'}{'supress_white_network_messages'};
                $is_lan_IP = 1;
                last;
              }
            }
            if ( $is_lan_IP == 0 ) {
              my %event_hash = ();
              $event_hash{'TIMESTAMP'} = $time_stamp;
              $event_hash{'IP'} = $ip_address;
              my $event_type = $config_hash{'proto'}{$proto}{$suffix_name}{'patterns'}{$suffix_number}{'type'};
              my $E_hash;
              if ( !defined $config_hash{'events'}{ $event_type } ) {
                logit ("!! Matched pattern type not defined! [proto-$proto]{$suffix_name\_$suffix_number} ($event_type)");
                last;
              }
              $E_hash = $config_hash{'events'}{ $event_type };
              $event_hash{'TYPE'} = $event_type;
              my $comment_inner_text = $config_hash{'events'}{$event_type}{'friendly_name'};
              # ha nem whitelist event és szerepel az unbanned_IP_hash táblában, akkor ismétlődőnek jelöljük
              if ( ( $config_hash{'events'}{$event_type}{'type'} eq 'B') && ( defined $unbanned_IP_hash{ $ip_address } ) ) {
                $event_hash{'REPEATED'} = $unbanned_IP_hash{ $ip_address }{'REPEATED'} +1;
                $comment_inner_text .= ' (repeated)';
              }
              $event_hash{'EXPIRE'} = $time_stamp + $config_hash{'basic'}{'trigger_time'};
              $event_hash{'COMMENT'} = "$date_time - ".$comment_inner_text." - username:[$username]" ;
              push @event_table , \%event_hash;
              print_to_queue "**debug_pattern_matching**       Event ($event_type) created!\n" if $debug_pattern_matching;
            }
            last; # foreach $suffix_number
          } # if match $config_hash{'proto'}{$proto}{$suffix_name}{'patterns'}{$suffix_number}{'pattern'}
        } # foreach {$suffix_number}
        print_to_queue "**debug_pattern_matching**   {$suffix_name} found_match: $found_match\n" if $debug_pattern_matching == 2;
        if ( $found_match ) { last; } # foreach $suffix_name
      } # foreach my $suffix_name
      print_to_queue "**debug_pattern_matching** {$proto} found_match: $found_match\n" if $debug_pattern_matching == 2;
      if ( $found_match ) { last; } # foreach $proto
    } # if match $config_hash{'proto'}{$proto}{'prefix'}
  } # foreach keys %{$config_hash{'proto'}}
}
$new_saved_filepos = tell $DATA;
close ($DATA);

#print_to_queue "========== new_saved_filepos $new_saved_filepos; saved_filepos $saved_filepos =========\n";
# kiírjuk a $LOGDATA fájlba az aktuális fájl pozíciót
if ( $new_saved_filepos != $saved_filepos ) {
  $LOGDATA = FileHandle->new($log_data_file, ">");
  if (!defined $LOGDATA) { die "Unable to open LOGDATA ($log_data_file) file: $!\n"; }
  seek $LOGDATA, 0, 0;
  print $LOGDATA "$saved_inode!$new_saved_filepos\n";
  close ($LOGDATA);
  $saved_filepos = $new_saved_filepos;
}

# töröljük az unbanned_IP_hash lejárt elemeit
foreach my $ip ( keys %unbanned_IP_hash) {
  if ( $current_time > $unbanned_IP_hash{$ip}{'EXPIRE'} + $unbanned_IP_remember_time ) {
    print_to_queue "Attack remember wait time reached for [$ip]\n";
    delete $unbanned_IP_hash{$ip};
  }
}

# lejárt elemeket keresünk a saved_IP_hash-ben
if ($found_matching_lines == 0 &&  $just_started == 0) {
  my $found_outdated = 0;
  foreach my $i ( sort { $saved_IP_hash{$a}->{'TIMESTAMP'} <=> $saved_IP_hash{$b}->{'TIMESTAMP'} } keys %saved_IP_hash) {
    if ( $current_time > $saved_IP_hash{$i}->{'EXPIRE'})  {
      $found_outdated++;
      print_to_queue 'Expired event found: ['.$saved_IP_hash{$i}->{'IP'}.'] - '.$saved_IP_hash{$i}->{'COMMENT'}."\n";
    }
  }
  if ( !$found_outdated )
    { sleep $sleep_time; next; } # minimum 1 sec, különben baj lesz a "$saved_IP_hash{'bb_'" soroknál
}
###############################################################################################################
############################### Itt megáll, ha nincs új elem ##################################################
###############################################################################################################
$just_started = 0;
{ # körönként törlendő dolgok ürítése
  # törlésre kerül a banned_IP_hash és az IP_type_hash tartalma
  # ezeket lentebb töltöm majd föl friss adatokkal (a lejárt elemeket kihagyva)
  for (keys %banned_IP_hash) { delete $banned_IP_hash{$_}; }
  for (keys %IP_type_hash) { delete $IP_type_hash{$_}; }
  # lejárt ismétlődés időzítők törlése: $ipt_messages_repeate_time
  foreach my $ipt_timers ( qw ( ipt_count_near_full ipt_count_should_lower ) ) {
    if ( defined $config_hash{'ipt'}{$ipt_timers} && $current_time > $config_hash{'ipt'}{$ipt_timers} + $ipt_messages_repeate_time ) {
      $config_hash{'ipt'}{$ipt_timers} = 0;
    }
  }
  # lejárt ismétlődés időzítők törlése: $ipt_errors_repeate_time
  foreach my $ipt_timers ( qw (ipt_list_tot_not_numeric ipt_count_not_numeric change_ipt_count_result ipt_recent_file_not_found) ) {
    if ( defined $config_hash{'ipt'}{$ipt_timers} && $current_time > $config_hash{'ipt'}{$ipt_timers} + $ipt_errors_repeate_time ) {
      $config_hash{'ipt'}{$ipt_timers} = 0;
    }
  }
}
########################### eredme'ny fa'jl (ssh_syslog_lines) kii'ra'sa #########################################
#print "---------- event_table -------------------------------\n". Dumper(@event_table) . "\n\n";
# Nincs duplikatum szures, mert elvileg nem is lehet!

my $KIMENET = FileHandle->new($outfile, ">" );
if (!defined $KIMENET) { die "Unable to open KIMENET ($outfile) file: $!\n"; }

$outdated_ittems = 0;
# végigmegyünk az @event_table összes elemén (régi + új)
for ( my $event_table_array_item_count = scalar( @event_table ); $event_table_array_item_count > 0; $event_table_array_item_count-- ) {
  # kivesszük az első elemet
  my $element = shift @event_table;
  if ( !defined $element ||
       !defined $element->{'IP'} ||
       !defined $element->{'TIMESTAMP'} ||
       !defined $element->{'EXPIRE'} ||
       !defined $element->{'TYPE'} ||
       !defined $element->{'COMMENT'} ) {
    print_to_queue 'An element of the @event_table corrupted. Dumping all elements ================ total '.@event_table."\n";
    print_to_queue ''.Dumper(@event_table)."\n";
    print_to_queue "\n================================\n";
    next;
  }
  my $Event_IP = $element->{'IP'};
  # ha lejárt az esemény ideje, töröljük a táblából és mentjük a neki megfelelő dátumú '$OLDDATA' fájlba
  if ( $current_time > $element->{'EXPIRE'} ) {
    my ($sec,$min,$hour,$mday,$mon,$year,$wday,$yday,$isdst) = localtime( $element->{'TIMESTAMP'} );
    $year = $year + 1900;
    $mon += 1;
    if( length( $mon ) == 1 ) { $mon = '0' . $mon; }
    if( length( $mday ) == 1 ) { $mday = '0' . $mday; }
    my $old_file_name = $work_dir.'/'.$year.'-'.$mon.'-'.$mday ;
    my $OLDDATA = FileHandle->new($old_file_name, ">>");
    if (!defined $OLDDATA) {
      print "Unable to open OLDDATA ($old_file_name) file: $!\n";
    } else {
      print $OLDDATA $element->{'TIMESTAMP'}.'!'.
                     $element->{'IP'}.'!'.
                     $element->{'TYPE'}.'!'.
                     $element->{'COMMENT'}."\n" ;
    }
    %{ $element } = (); ######## felszabaditjuk a hasht ##
    $outdated_ittems++;
    next; # for(each) @event_table ...
  }
  # visszatesszük a végére ha még aktuális
  push @event_table, $element;
  # IP_type_hash feltöltése a még aktuális eseményekkel (körönként üres)
  put_in_IP_type_hash($element);
  # ha ban az esemény típusa, akkor kiírjuk
  # a whitelist elemekből csak 1-et írunk ki később a fájl végére, hogy ne kelljen keresgélni őket
  if ( $config_hash{'events'}{ $element->{'TYPE'} }{'type'} eq 'B' ) {
    # ssh_syslog_lines.txt.new
    print $KIMENET $element->{'TIMESTAMP'}.'!'.
                   $element->{'IP'}.'!'.
                   $element->{'TYPE'}.'!'.
                   $element->{'COMMENT'}."\n" ;
  }
}
# whitelist elemek kiírása (1 elem / IP)
foreach my $ip ( keys %IP_type_hash) {
  foreach my $event_type ( keys %{ $IP_type_hash{$ip} } ) {
    if ( defined $config_hash{'events'}{$event_type} && $config_hash{'events'}{$event_type}{'type'} eq 'W' ) {
#      print_to_queue "White IP: $ip\n";
      print $KIMENET $IP_type_hash{$ip}{$event_type}{'TIMESTAMP'}.'!'.$ip.'!'.$event_type.'!'.$IP_type_hash{$ip}{$event_type}{'COMMENT'}."\n" ;
    }
  }
}
close ($KIMENET);
if (-f $infile.'.last' ) {
  if (!unlink ($infile.'.last') ) {
    die "Unable to delete [$infile.last] file: $!\n";
  }
}
if (-f $infile ) {
  if (! rename($infile, $infile.'.last') ) {
    die "Unable to rename [$infile] file to [$infile.last]: $!\n";
  }
}
if (! rename($outfile, $infile) ) {
  die "Unable to rename [$outfile] file to [$infile]: $!\n";
}
print_to_queue '######## '.scalar localtime()." ------ Last read byte of the logfile: $saved_filepos || $found_matching_lines new line and $outdated_ittems outdated item(s) were found.\n";
$outdated_ittems = 0;

################################# IP_type_hash takarítása ############################################
################################# saved_IP_hash frissítése, feltöltése ###############################
#print Dumper(@event_table) . "\n-----------------------------------------------------------------------\n";
#print_to_queue "---------- IP_type_hash -------------------------------\n". Dumper(%IP_type_hash) . "\n\n";
#print_to_queue "---------- saved_IP_hash -------------------------------\n". Dumper(%saved_IP_hash) . "\n\n";

$new_ip = 0;
$new_ittems = 0;
$banned_IP_count = 0;

foreach my $ip ( keys %IP_type_hash) {
  my $debug_iptypehash = 0;
  print_to_queue "**debug_iptypehash** $ip\n" if $debug_iptypehash;
  my $is_uniq_ip = 1;
# ha volt az adott IP-ről whitelist esemény akkor törlünk minden bant
  foreach my $event_type ( keys %{ $IP_type_hash{$ip} } ) {
    if ( defined $config_hash{'events'}{$event_type} && $config_hash{'events'}{$event_type}{'type'} eq 'W' ) {
      print_to_queue "**debug_iptypehash**   $ip has whitelist event. Deleting everithing else.\n" if $debug_iptypehash;
      for (keys %{ $IP_type_hash{$ip} } ) {
        if ( defined $config_hash{'events'}{$_} && $config_hash{'events'}{$_}{'type'} eq 'B' ) {
          $IP_type_hash{$ip}{$_} = (); ######## felszabaditjuk a hasht ##
          delete $IP_type_hash{$ip}{$_};
        }
      }
    }
  }
# minden esemény típust megnézünk az adott IP-n belül
  foreach my $event_type ( keys %{ $IP_type_hash{$ip} } ) {
    my $E_hash = $config_hash{'events'}{$event_type};
    my $event_time = $E_hash->{'ban_time'};
    # ha nem elérte el a szükséges esemény számot a banhoz akkor töröljük a teljes esemény típust
    if ( $IP_type_hash{$ip}{$event_type}{'count'} < $E_hash->{'trigger_count'}  ) {
      print_to_queue "**debug_iptypehash**   $ip has not enough '$event_type' type events (".$IP_type_hash{$ip}{$event_type}{'count'}.'<'.$E_hash->{'trigger_count'}."). Deleting it.\n" if $debug_iptypehash;
      %{$IP_type_hash{$ip}{$event_type}} = ();
      delete $IP_type_hash{$ip}{$event_type};
      next;
    }
    # ha az utolsó bejegyzés régebbi, mint az eseménhez tartozó ban_time akkor töröljük a teljes esemény típust
    if ( $current_time > $IP_type_hash{$ip}{$event_type}{'TIMESTAMP'} + $event_time ) {
      print_to_queue "**debug_iptypehash**   $ip has no recent '$event_type' type events (".($IP_type_hash{$ip}{$event_type}{'TIMESTAMP'} + $event_time)."<$current_time). Deleting it.\n" if $debug_iptypehash;
      %{$IP_type_hash{$ip}{$event_type}} = ();
      delete $IP_type_hash{$ip}{$event_type};
      next;
    }
  }
  my @event_types = keys %{ $IP_type_hash{$ip} };
#  print_to_queue '@event_types='."@event_types\n";
  if ( scalar @event_types == 0 ) {
    # nem maradt esemény típus, töröljük az IP-t
    print_to_queue "**debug_iptypehash**   $ip has no events remained, deleting.\n" if $debug_iptypehash;
    delete $IP_type_hash{$ip};
    next;
  }
# végig nézzük az adott IP minden megmaradt esemény típusát
  foreach my $event_type ( @event_types ) {
    my $is_uniq_type = 1;
    my $is_repeated_event = '';
    my $E_hash = $config_hash{'events'}{$event_type};
    my $event_time = $E_hash->{'ban_time'};
    if ( $IP_type_hash{$ip}{$event_type}{'REPEATED'} ) {
      $is_repeated_event = ' repeated';
      $event_time = $E_hash->{'ban_time'} * $ban_time_multiplier * $IP_type_hash{$ip}{$event_type}{'REPEATED'};
    }
    my $expire_time = $IP_type_hash{$ip}{$event_type}{'TIMESTAMP'} + $event_time;
# végignézzük az összes eddigi mentett IP-t. Ha van ugyanilyen IP és esemény, akkor frissítjük az időt és lejáratot
    foreach my $s_id (keys %saved_IP_hash) {
      if ( ($saved_IP_hash{$s_id}{'IP'} eq $ip) && ($saved_IP_hash{$s_id}{'TYPE'} == $event_type) ) {
        print_to_queue "**debug_iptypehash**   $ip event '$event_type' is not uniq. Found at saved_IP_hash{$s_id}\n" if $debug_iptypehash;
        $is_uniq_type = 0;
        $is_uniq_ip = 0;
        # ismétlődő eseménynél a $saved_IP_hash lejárata akár sokkal több is lehet, mint az $IP_type_hash-beli!
        if ( $expire_time > $saved_IP_hash{$s_id}{'EXPIRE'} ) {
          $saved_IP_hash{$s_id}{'TIMESTAMP'} = $IP_type_hash{$ip}{$event_type}{'TIMESTAMP'};
          $saved_IP_hash{$s_id}{'EXPIRE'} = $expire_time;
          $saved_IP_hash{$s_id}{'COMMENT'} = $IP_type_hash{$ip}{$event_type}{'COMMENT'};
        }
      }
    }
# ha még nem volt ilyen IP mentve, akkor mentjük és naplózzuk az új eseményt (ban/whitelist)
    if ($is_uniq_type) {
      $new_ittems++;
      $saved_IP_hash{'bb_'.$current_time.'_'.$new_ittems} = \%{ $IP_type_hash{$ip}{$event_type} };
      delete $saved_IP_hash{'bb_'.$current_time.'_'.$new_ittems}{'count'};
      $saved_IP_hash{'bb_'.$current_time.'_'.$new_ittems}{'IP'} = $ip;
      $saved_IP_hash{'bb_'.$current_time.'_'.$new_ittems}{'TYPE'} = $event_type;
      $saved_IP_hash{'bb_'.$current_time.'_'.$new_ittems}{'EXPIRE'} = $expire_time;
      if ( $config_hash{'events'}{$event_type}{'type'} eq 'W' ) {
        my $tmp_user = 'N/A';
        if ( $IP_type_hash{$ip}{$event_type}{'COMMENT'} =~ m/ - username:\[(.*)\]/ ) {
          $tmp_user = $1;
        }
        logit ("[a] New event [$config_hash{'events'}{$event_type}{'friendly_name'}] for user [$tmp_user] from IP [$ip] added to event spool for $event_time seconds");
      } elsif ( $config_hash{'events'}{$event_type}{'type'} eq 'B' ) {
        if ($is_uniq_ip) { $new_ip++; }
        my $commited = ipt_recent('add', $ip);
        logit ("[a] New$is_repeated_event event [$config_hash{'events'}{$event_type}{'friendly_name'}] from IP [$ip] added to iptables recent list for $event_time seconds") if $commited;
      } else {
        logit ("[a] Unknown event type [$config_hash{'events'}{$event_type}{'type'}]");
      }
    }
  }
}

################################# saved_IP_hash kiírása (eredmény fájl) #######################################
#print "---------- saved_IP_hash -------------------------------\n". Dumper(%saved_IP_hash) . "\n\n";
my $SAVEDIP_NEWFILE = FileHandle->new($saved_IP_newfile, ">");
if (!defined $SAVEDIP_NEWFILE) { die "Unable to open BANOUTFILE ($saved_IP_newfile) file: $!\n" };

# időrendben végigmegyünk a saved_IP_hash elemein
foreach my $i ( sort { $saved_IP_hash{$a}->{'TIMESTAMP'} <=> $saved_IP_hash{$b}->{'TIMESTAMP'} ||
                       $saved_IP_hash{$a}->{'IP'} cmp $saved_IP_hash{$b}->{'IP'} ||
                       $saved_IP_hash{$a}->{'TYPE'} <=> $saved_IP_hash{$b}->{'TYPE'}
                     } keys %saved_IP_hash) {
# ha lejárt az ideje, töröljük az eseményt (egy IP-hez lehet több esemény fajta is!!!)
  if ( $current_time > $saved_IP_hash{$i}->{'EXPIRE'} ) {
    $outdated_ittems++;
# mentjük vagy frissítjük az $unbanned_IP_hash táblában az adatokat
    my $Event_IP = $saved_IP_hash{$i}->{'IP'};
    if ( defined $unbanned_IP_hash{$Event_IP} ) {
      $unbanned_IP_hash{$Event_IP}{'REPEATED'} += 1;
      my $current_delta = $saved_IP_hash{$i}->{'EXPIRE'} - $saved_IP_hash{$i}->{'TIMESTAMP'};
      my $old_delta = $unbanned_IP_hash{$Event_IP}{'EXPIRE'} - $unbanned_IP_hash{$Event_IP}{'TIMESTAMP'};
      if ( $current_delta > $old_delta ) {
        $unbanned_IP_hash{$Event_IP}{'TIMESTAMP'} = $saved_IP_hash{$i}->{'TIMESTAMP'};
        $unbanned_IP_hash{$Event_IP}{'EXPIRE'} = $saved_IP_hash{$i}->{'EXPIRE'};
      }
    } else {
      my $event_type = $saved_IP_hash{$i}->{'TYPE'};
      if ( $config_hash{'events'}{$event_type}{'type'} eq 'B') {
        $unbanned_IP_hash{$Event_IP}{'TIMESTAMP'} = $saved_IP_hash{$i}->{'TIMESTAMP'};
        $unbanned_IP_hash{$Event_IP}{'EXPIRE'} = $saved_IP_hash{$i}->{'EXPIRE'};
        $unbanned_IP_hash{$Event_IP}{'REPEATED'} = $saved_IP_hash{$i}->{'REPEATED'} || 0;
      }
    }
    $IP_type_hash{$Event_IP}{'expired'}{ $config_hash{'events'}{$saved_IP_hash{$i}->{'TYPE'}}{'friendly_name'} } = 1; # feljegyezzük, mint potenciálisan törölhető bant
    # magát az eseményt lehet törölni, mert az lejárt
    %{$saved_IP_hash{$i}} = (); delete $saved_IP_hash{$i};
  } else {
# ha érvényes még, és ban akkor mentjük a banned_IP_hash táblába
    if ( $config_hash{'events'}{ $saved_IP_hash{$i}->{'TYPE'} }{'type'} eq 'B') {
      $IP_type_hash{ $saved_IP_hash{$i}->{'IP'} }{'valid'} = 1; # feljegyezzük, mint nem törölhető bant
      if ( !defined $banned_IP_hash{ $saved_IP_hash{$i}->{'IP'} } ) {
        $banned_IP_hash{ $saved_IP_hash{$i}->{'IP'} }{'TIMESTAMP'} = $saved_IP_hash{$i}->{'TIMESTAMP'};
        $banned_IP_count++;
      }
    }
    print $SAVEDIP_NEWFILE $saved_IP_hash{$i}->{'TIMESTAMP'}.'!'.
                           $saved_IP_hash{$i}->{'IP'}.'!'.
                           ( defined $saved_IP_hash{$i}->{'REPEATED'}?'R':'').$saved_IP_hash{$i}->{'TYPE'}.( defined $saved_IP_hash{$i}->{'REPEATED'} ? '-'.$saved_IP_hash{$i}->{'REPEATED'}.'-'.$saved_IP_hash{$i}->{'EXPIRE'} :'').'!'.
                           $saved_IP_hash{$i}->{'COMMENT'}."\n" ;
  }
}
close ($SAVEDIP_NEWFILE);
# Töröljük azokat a banokat amiket lehet, a lejártakat csak naplózzuk
foreach my $ip ( keys %IP_type_hash) {
  if (!defined $IP_type_hash{$ip}{'expired'}) {
    delete $IP_type_hash{$ip}{'valid'};
    next;
  }
  # ha nincs több esemény ami miatt 'banolva' van akkor töröljük a bant
  if ( !defined $IP_type_hash{$ip}{'valid'} ) {
    foreach my $friendly_name ( keys %{ $IP_type_hash{$ip}{'expired'} } ) {
      my $commited = ipt_recent('remove', $ip );
      logit ("[b] Event [$friendly_name] expired from IP [$ip] and deleted from iptables recent list") if $commited;
    }
  # ha van még, akkor csak értesítés megy, hogy lejárt egy ban
  } else {
    foreach my $friendly_name ( keys %{ $IP_type_hash{$ip}{'expired'} } ) {
      logit ("[b] Event [$friendly_name] expired from IP [$ip] and deleted from event spool");
    }
    logit ("[b] other bans are still exist for this IP so it will remain banned!");
  }
  delete $IP_type_hash{$ip}{'expired'};
  delete $IP_type_hash{$ip}{'valid'};
  delete $IP_type_hash{$ip} if ( scalar (keys %{ $IP_type_hash{$ip} } ) == 0 );
}
print_to_queue "$new_ittems new event triggered for $new_ip new IP and $outdated_ittems ban(s) expired. | Currently $banned_IP_count ban(s) exists.\n";

if (-f $saved_IP_file.'.last' ) {
  if (!unlink ($saved_IP_file.'.last') ) {
    die "Unable to delete [$saved_IP_file.last] file $!\n";
  }
}
if (-f $saved_IP_file ) {
  if (! rename($saved_IP_file, $saved_IP_file.'.last') ) {
    die "Unable to rename [$saved_IP_file] file to [$saved_IP_file.last]: $!\n";
  }
}
if (! rename($saved_IP_newfile, $saved_IP_file) ) {
  die "Unable to rename [$saved_IP_newfile] file to [$saved_IP_file]: $!\n";
}

if ( check_iptables_match_file_path()  ) {
  my %ipt_IP_hash;
  my $current_banspool = '';
  my $banspool_count = 0;
  my $ipt_content = `cat $iptables_match_file_path`;
  my $ipt_count = `cat $iptables_match_file_path | wc -l`;
  my $ipt_list_tot = `cat /sys/module/$iptables_recent_dir/parameters/ip_list_tot`;
  chomp $ipt_list_tot;
  if ( $ipt_list_tot !~ m/^(\d+)$/ ) {
    if ( !defined $config_hash{'ipt'}{'ipt_list_tot_not_numeric'} || $config_hash{'ipt'}{'ipt_list_tot_not_numeric'} == 0 ) {
      $config_hash{'ipt'}{'ipt_list_tot_not_numeric'} = $current_time;
      logit ("ipt_list_tot is not numeric: $ipt_list_tot");
    }
    $ipt_list_tot = 0;
  }
  chomp $ipt_count;
  if ( $ipt_count !~ m/^(\d+)$/ ) {
    if ( !defined $config_hash{'ipt'}{'ipt_count_not_numeric'} || $config_hash{'ipt'}{'ipt_count_not_numeric'} == 0 ) {
      $config_hash{'ipt'}{'ipt_count_not_numeric'} = $current_time;
      logit ("ipt_count is not numeric: $ipt_count");
    }
    $ipt_count = 0;
  }
 # print_to_queue ("ipt_content: \n" . $ipt_content);
  for (split /^/, $ipt_content) {
    if ( $_ =~ m/^src=(.*) ttl: / ) {
      $ipt_IP_hash{ $1 } = '1';
    } else {
      logit ("iptables recent line didn't match /^src=(.*) ttl: /: $_");
    }
  }
  if ( $ipt_count + 20 > $ipt_list_tot ) {
    change_ip_list_tot(\$ipt_list_tot, $ipt_count, $ipt_list_tot+20 );
  }
  if ( $ipt_list_tot >= 120 && $ipt_count < $ipt_list_tot - 40 ) {
    change_ip_list_tot(\$ipt_list_tot, $ipt_count, $ipt_list_tot-20);
  }
  foreach my $ip ( sort { $banned_IP_hash{$b}->{'TIMESTAMP'} <=> $banned_IP_hash{$a}->{'TIMESTAMP'} } keys %banned_IP_hash ) {
    $current_banspool .= $ip.' ' if $max_current_banspool_IP_to_syslog > $banspool_count;
    $banspool_count++;
    if ( $ipt_list_tot != 0 && $ipt_list_tot + 20 < $banspool_count ) {
      change_ip_list_tot(\$ipt_list_tot, $banspool_count, $banspool_count+20);
    }
    if ( !defined $ipt_IP_hash{ $ip } ) {
      
      my $add_result = ipt_recent('add', $ip);
      if ( $add_result ) {
        logit ("[c1] IP: $ip disappeared from iptables recent file ($iptables_match_file_path). It was reinserted.");
      } else {
        logit ("[c1] IP: $ip disappeared from iptables recent file ($iptables_match_file_path). Reinsertion failed.");
      }
    } else {
      delete $ipt_IP_hash{ $ip };
    }
  }
  if ( $max_current_banspool_IP_to_syslog != 0 && ($new_ittems || $outdated_ittems) ) {
    logit ("[c] Current ban spool is [$current_banspool]");
  }
  foreach my $ip ( keys %ipt_IP_hash ) {
    my $remove_result = ipt_recent('remove', $ip);
      if ( $remove_result ) {
        logit ("[c2] unbanned IP: $ip appeared in iptables recent file ($iptables_match_file_path). It was deleted.");
      } else {
        logit ("[c2] unbanned IP: $ip appeared in iptables recent file ($iptables_match_file_path). Deletion failed.");
      }
  }
} else {
  logit ("[aa] configured 'iptables_recent_name' didn't exist: /proc/net/(ipt_recent|xt_recent)/$iptables_recent_name didn't exist!");
}

sleep $sleep_time; next; # minimum 1 sec, különben baj lesz a "$saved_IP_hash{'bb_'" soroknál
}

print "\n";
unlink( $flag );
exit 0;
