#!/usr/bin/perl -w
########################################################

use strict;
use warnings;
use utf8;
use Encode qw( from_to encode decode );
use FileHandle;
use File::Basename;
use Sys::Syslog;
use Data::Dumper;
use POSIX;
use Config::IniFiles;
use Mail::Sendmail;
#use SendMail 2.11;

my $config_file = "/etc/wwm.cfg";
my $local_config_file = "/etc/wwm.local.cfg";
my $prg = basename( $0 );
my $prg2 = '';
my $exit_code = 0;
my $restart_count_limit = 0;
my $prg_dir = '';
my $work_dir = '';
my $logfile = '';
my $log_data_file = '';
my $mail_from = '';
my $mail_to = '';
my $mail_subject = '';
my $mail_last_lines = 0;
my @last_lines;

sub _from_wide( $ ) {
  my $p0 = shift;
  my $p1 = '';
  foreach my $item ( split( //, $p0 ) ) {
    $p1 = $p1 . chr( ord( $item ) & 255 );
    $p1 = $p1 . chr( ( ord( $item ) / 256 ) & 255 );
  }
  from_to( $p1, 'utf16-le', 'utf8' );
  return $p1;
}

sub _to_utf8( $ ) {
  my $p0 = shift;
  my $p1 = '';
  my $p2 = '';
  my $p3 = '';
  my $p4 = '';
  foreach my $chars ( 'utf8', 'iso-8859-2', 'iso-8859-1' ) {
    $p1 = decode( $chars, $p0 );
    $p2 = encode( 'utf8', $p1 );
    $p3 = decode( 'utf8', $p2 );
    $p4 = encode( $chars, $p3 );
    if ( $p1 eq $p3 and $p2 ne $p3 and $p4 eq $p0 ) {
      if (    ( $p4 ne $p2 and $chars =~ /^iso/ )
           or ( $p4 ne $p3 and $chars =~ /^utf/ ) ) {
        return $p2;
      }
    }
  }
  return $p0;
}

sub logit { ##################### sub print - Begin
  my( $msg ) = @_;
  if( length( $msg ) == 0 ) {
    return 1;
  }
  $msg = _to_utf8( _from_wide( $msg ) );
  print "###-Logit-### $msg\n";
#  setlogsock( 'unix' ); # http://perldoc.perl.org/Sys/Syslog.html  [THE RULES OF SYS::SYSLOG]
  openlog( $prg, 'pid,ndelay,nowait', 'user' );
  syslog( 'info', $msg );
  closelog(  );
  return 1;
} ################################ sub print - End

sub EncodeToMIME( @ ) {
  my %args = ( String => undef, Charset => undef, @_ );
  my $value = $args{'String'};
  return $value unless $value;
  my $charset  = $args{'Charset'} || 'utf-8';
  my $encoding = 'B';
  my $max      = int( ( ( 75 - length( '=?' . $charset . '?' . $encoding . '?' . '?=' ) ) * 3 ) / 4 );
  $max = int( $max / 3 ) * 3;
  chomp $value;

  if ( $max <= 0 ) {
    return ( $value );
  }
  return ( $value ) if $value =~ /^(?:[\t\x20-\x7e]|\x0D*\x0A[ \t])+$/s;
  $value =~ s/\s+$//;
  Encode::_utf8_on( $value ) unless Encode::is_utf8( $value );
  my ( $tmp, @chunks ) = ( '', () );
  while ( length $value ) {
    my $char = substr( $value, 0, 1, '' );
    my $octets = encode( $charset, $char );
    if ( length( $tmp ) + length( $octets ) > $max ) {
      push @chunks, $tmp;
      $tmp = '';
    }
    $tmp .= $octets;
  }
  push @chunks, $tmp if length $tmp;
  $value = join( "\n ", map encode_mimeword( $_, $encoding, $charset ), @chunks );
  return ( $value );
}
sub MIMEEncodeString( $ ) {
  return EncodeToMIME( String => $_[0], Charset => 'utf-8' );
}
sub mail_to {
  my $mail_body = shift;
  my %mail = ( To      => $mail_to,
            From    => $mail_from,
            Message => _to_utf8( $mail_body ),
            Subject => MIMEEncodeString( _to_utf8( $mail_subject . ' ('.scalar localtime(). ')' ) ),
           );
  if ( sendmail %mail ) {
    print( "Mail sent to: $mail_to\n" );
    print( "----------------------\n");
    print( $Mail::Sendmail::log ."\n");
    print( "----------------------\n");
    return 0;
  } else {
    print( "Error sending mail to $mail_to:\n" );
    print( "----------------------\n");
    print( $Mail::Sendmail::error ."\n");
    print( "----------------------\n");
    return 1;
  }
}

sub read_value_from_config {
  my( $config, $section, $param ) = @_;
  my @tmp = $config->val( $section,  $param );
  if (@tmp > 1) {
    print "WARNING! Multiline config parameter [$section][$param] merged to a single line:\n";
    print "[".join('', @tmp)."]\n";
    return join('', @tmp);
  }
  return $tmp[0];
}

sub read_config {
  my %new_config_hash;
  my @configs = ( { type => 'system', file => $config_file }, { type => 'local', file => $local_config_file } );
  foreach ( @configs ) {
    if ( -f $_->{'file'} ) {
      my $config_type = $_->{'type'};
      my $config = Config::IniFiles->new( -file => $_->{'file'} );
      print "################# config_file = $_->{'file'} ##############################\n";
#print "---------- $config_file -------------------------------\n". Dumper($config) . "\n\n";
      if (! defined $config ) {
        print STDERR "Config file '$_->{'file'}' has errors:\n$@";
        exit 51;
      }
      my $sect = 'basic';
      foreach my $param ( qw ( logfile work_dir ) ) {
        my $tmp = read_value_from_config ($config, $sect, $param );
        if ( defined $tmp ) {
          $new_config_hash{$sect}{$param} = $tmp;
          print "$param = $tmp\n";
        } elsif ( $config_type eq 'system' ) {
          print STDERR "Mandatory parameter [$param] missing from config file '$sect' section!";
          exit 50;
        }
      }
      $sect = 'runner';
      foreach my $param ( qw ( main_program_name main_program_dir main_program_restart_limit mail_from mail_to mail_subject mail_append_last_n_lines ) ) {
        my $tmp = read_value_from_config ($config, $sect, $param );
        if ( defined $tmp ) {
          $new_config_hash{$sect}{$param} = $tmp;
          print "$param = $tmp\n";
        } elsif ( $config_type eq 'system' ) {
          print STDERR "Mandatory parameter [$param] missing from config file '$sect' section!";
          exit 50;
        }
      }
    } else {
      print STDERR "Config file ($_->{'file'}) did not exist!\n";
      exit 50;
    }
  }
  print "read_config Done.\n";
  $logfile              = $new_config_hash{'basic'}{'logfile'};
  $work_dir             = $new_config_hash{'basic'}{'work_dir'};
  $log_data_file        = $work_dir . "/logdata.txt";
  $prg2                 = $new_config_hash{'runner'}{'main_program_name'};
  $prg_dir              = $new_config_hash{'runner'}{'main_program_dir'};
  $restart_count_limit  = $new_config_hash{'runner'}{'main_program_restart_limit'};
  $mail_from            = $new_config_hash{'runner'}{'mail_from'};
  $mail_to              = $new_config_hash{'runner'}{'mail_to'};
  $mail_subject         = $new_config_hash{'runner'}{'mail_subject'};
  $mail_last_lines      = $new_config_hash{'runner'}{'mail_append_last_n_lines'};

  {
    $Data::Dumper::Purity = 1;
    $Data::Dumper::Sortkeys = sub {
      my @KEYS = keys %{$_[0]};
      return [@KEYS] unless defined $KEYS[0];
      $KEYS[0] =~ m/^\d+$/ ? return [sort {$a <=> $b} @KEYS ] : return [sort {$a cmp $b} @KEYS ] 
    };
    print '' . Data::Dumper->Dump([{ %new_config_hash }], ['config_hash']);
  }
  return 1;
}
######################### start ####################
$| = 1;
read_config();
print "prg <$prg> | prg_dir <$prg_dir> | prg2 <$prg2> | work_dir <$work_dir>\n";
my $tmpstr = shift || '';
if ( $tmpstr eq '--testmail' ) {
  print( "Got '--testmail' parameter.\n" );
  print( "Sending mail from '$mail_from' to '$mail_to' with subject: '$mail_subject'\n" );
  $tmpstr = mail_to('Teszt');
  exit $tmpstr;
}

my $flag = "/var/run/$prg"; #print "$flag\n";

if (-f $flag ) {
  print STDERR ( "Already running\n" );
  exit 0;
}
my $F1;
die "Semaphore file ($flag) creation failed: $!" unless open $F1, '>', $flag;
die "Semaphore file ($flag) close failed: $!"unless close $F1;

my ($current_time, $start_time, $delta);
my ($day, $hour, $minute, $second);
my $restart_count = 0;

if (! -f "$prg_dir$prg2") {
  unlink( $flag );
  die "Program file ($prg_dir$prg2) didn't exist:\n$!\n"
}
$tmpstr = "ps awxf | grep '$prg2' | grep perl | grep -v grep | awk '{print " . '$1' . "}'";
my $run_check = `$tmpstr`;
if ( $run_check ne '' ) {
  unlink( $flag );
  die "Program ($prg2) already running!\n"
}
my $LOGFILE = FileHandle->new( $work_dir.'/output.txt', ">");
if (!defined $LOGFILE) { 
  unlink( $flag );
  die "Unable to open $work_dir/output.txt file: $!\n" 
};
close($LOGFILE);
while (1) {
  undef(@last_lines);
  $start_time = time();
  open (PRG, "$prg_dir$prg2 2>&1 |");
  while ( <PRG> )  {
    $LOGFILE = FileHandle->new( $work_dir.'/output.txt', ">>");
    $_ = _to_utf8( _from_wide( $_ ) );
    print $LOGFILE "$_";
    close($LOGFILE);
    print 'child |'.$_;
    if ( $mail_last_lines ) {
      push @last_lines, $_;
      shift @last_lines if scalar @last_lines > $mail_last_lines;
    }
  }
  close (PRG);
  $exit_code = $? >> 8; # ${^CHILD_ERROR_NATIVE} >> 8;
  printf "###### !! program exited with exit code $exit_code\n";
#  unlink( $flag ); exit; # ########################################################################
  $current_time = time();
  $delta = $current_time - $start_time;
  $day = floor( $delta / 86400 );
  $hour   = floor( ($delta - ($day * 86400) ) / 3600 );
  $minute = floor( ($delta - ($day * 86400) - ($hour * 3600) ) / 60 );
  $second =      ( ($delta - ($day * 86400) - ($hour * 3600) - ($minute * 60) ) );
  $LOGFILE = FileHandle->new( $work_dir.'/output.txt', ">>");
  print $LOGFILE "################# ".scalar localtime();
  $tmpstr = sprintf "$prg2 exited ($exit_code) after $day day %02d:%02d:%02d runtime",$hour,$minute,$second;
  print $LOGFILE "################# $tmpstr\n";
  if ( $exit_code == 100 ) {
    print $LOGFILE "################# Manual exit. Program will halt.";
    logit( $tmpstr.". This is a manual exit. Program will halt." );
  } elsif ( $exit_code == 101 ) {
    print $LOGFILE "################# todo restart command.";
    logit( $tmpstr.". This is a manual restart. Program will restart." );
  } elsif ( $exit_code == 200 ) {
    $tmpstr .= "\nBanning rule did not exist in (any) iptables filter chain\n";
    $tmpstr .= "There is no point staying resident if no banning rule present! Quiting.";
    mail_to($tmpstr);
    unlink( $flag );
    exit 0;
  } else {
    $tmpstr .= "restart_count: $restart_count | restart_count_limit: $restart_count_limit\n";
    logit( $tmpstr );
  }
  close($LOGFILE);
  unlink( "/var/run/$prg2" );
  next if $exit_code == 101; # restart by command in todo file
  last if $exit_code == 100; # exit by command in todo file
  last if $exit_code == 200; # no banning rule present
  sleep (1);
  if ( $delta > 60 ) {
    # ha legalább 1 órát futott, akkor töröljük az újraindítás számlálót
    if ( $hour > 1 || $day > 1 ) { $restart_count = 0; }
    print "################# program runned at least an hour without error.\n";
    print "################# restart_count cleared.\n";
  }
  $restart_count++;
  if ( $restart_count == 1 ) {
    logit "Program will be lauched automatically!";
    if ( $mail_last_lines ) {
      $tmpstr .= "\n\nLast $mail_last_lines lines of program output:\n". join( '', @last_lines )."\n";
    }
    mail_to($tmpstr);
    next;
  }
  if ( $restart_count > $restart_count_limit ) {
    print "################# maximum restart limit reached.\n";
    logit "maximum restart limit reached. Program will not be lauched automatically!";
    $tmpstr .= "maximum restart limit reached. Program will not be lauched automatically!\n";
    print "################# sending mail\n";
    if ( $mail_last_lines ) {
      $tmpstr .= "\n\nLast $mail_last_lines lines of program output:\n". join( '', @last_lines )."\n";
    }
    mail_to($tmpstr);
    last;
  }
  else {
    $tmpstr .= "Program will be lauched automatically!\n";
    $tmpstr .= "To prevent program failures caused by logline processing errors,\n";
    $tmpstr .= "unprocessed lines will be skipped!\n";
    logit "Program will be lauched automatically! To prevent program failures caused by logline processing errors, unprocessed lines will be skipped!";
    print "################# sending mail\n";
    if ( $mail_last_lines ) {
      $tmpstr .= "\n\nLast $mail_last_lines lines of program output:\n". join( '', @last_lines )."\n";
    }
    mail_to($tmpstr);
    
    if (-f $logfile) {
      my $DATA = FileHandle->new($logfile);
      if (!defined $DATA) {
        logit "Unable to open DATA ($logfile) file: $!"; 
        last;
      }
      my $log_inode = (stat $DATA)[1];
      my $log_size = (-s $logfile);
      if ( -f $log_data_file ) {
        my $LOGDATA = FileHandle->new($log_data_file, ">");
        if (!defined $LOGDATA) {
          logit "Unable to open LOGDATA ($log_data_file) file: $!";
          last;
        } else {
          seek $LOGDATA, 0, 0;
          print $LOGDATA "$log_inode!$log_size\n";
          close ($LOGDATA);
        }
      } else {
        logit "LOGDATA ($log_data_file) file didn't exist: $!";
        last;
      }
    } else {
      logit "DATA ($logfile) file didn't exist: $!";
      last;
    }
  }
}

print "\n";
unlink( $flag );
exit 0;

